from collections.abc import Callable, Iterable, Sequence
from itertools import islice, pairwise
from typing import Any

import numpy as np
import scipy
import sympy
from scipy.integrate import dblquad, nquad, quad, tplquad
from scipy.stats._qmc import Sobol
from sympy import solve
from sympy.stats import density, pspace
from sympy.stats.rv import is_random
from sympy.utilities import lambdify

from ..._types import FloatNDArray
from .._convex_hull import _compute_convex_hull
from ..common import classification_threshold
from .common import MetricFn, RateFn, ThresholdFn, _check_parameters_decorator

_sympy_dist_to_scipy: dict[
    sympy.stats.crv_types.SingleContinuousDistribution | sympy.stats.drv_types.SingleDiscreteDistribution,
    scipy.stats.rv_continuous | scipy.stats.rv_discrete,
] = {
    sympy.stats.crv_types.ArcsinDistribution: scipy.stats.arcsine,
    sympy.stats.crv_types.BetaDistribution: scipy.stats.beta,
    sympy.stats.crv_types.BetaPrimeDistribution: scipy.stats.betaprime,
    sympy.stats.crv_types.ChiDistribution: scipy.stats.chi,
    sympy.stats.crv_types.ChiSquaredDistribution: scipy.stats.chi2,
    # sympy.stats.crv_types.Erlang: scipy.stats.chi2
    # Erlang internally calls GammaDistribution so should be supported
    sympy.stats.crv_types.ExGaussianDistribution: scipy.stats.exponnorm,
    sympy.stats.crv_types.ExponentialDistribution: scipy.stats.expon,
    sympy.stats.crv_types.FDistributionDistribution: scipy.stats.f,
    sympy.stats.crv_types.GammaDistribution: scipy.stats.gamma,
    sympy.stats.crv_types.GammaInverseDistribution: scipy.stats.invgamma,
    sympy.stats.crv_types.LaplaceDistribution: scipy.stats.laplace,
    sympy.stats.crv_types.LogisticDistribution: scipy.stats.logistic,
    sympy.stats.crv_types.LogNormalDistribution: scipy.stats.lognorm,
    sympy.stats.crv_types.LomaxDistribution: scipy.stats.lomax,
    sympy.stats.crv_types.MaxwellDistribution: scipy.stats.maxwell,
    sympy.stats.crv_types.MoyalDistribution: scipy.stats.moyal,
    sympy.stats.crv_types.NakagamiDistribution: scipy.stats.nakagami,
    sympy.stats.crv_types.NormalDistribution: scipy.stats.norm,
    sympy.stats.crv_types.PowerFunctionDistribution: scipy.stats.powerlaw,
    sympy.stats.crv_types.StudentTDistribution: scipy.stats.t,
    sympy.stats.crv_types.TrapezoidalDistribution: scipy.stats.trapezoid,
    sympy.stats.crv_types.TriangularDistribution: scipy.stats.triang,
    sympy.stats.crv_types.UniformDistribution: scipy.stats.uniform,
    sympy.stats.crv_types.GaussianInverseDistribution: scipy.stats.invgauss,
}

_sympy_dist_to_scipy_params: dict[
    sympy.stats.crv_types.SingleContinuousDistribution | sympy.stats.drv_types.SingleDiscreteDistribution,
    Callable[..., dict[str, float]],
] = {
    sympy.stats.crv_types.ExponentialDistribution: lambda rate: {'loc': 0, 'scale': 1 / rate},
    sympy.stats.crv_types.GammaDistribution: lambda k, theta: {'a': k, 'scale': theta},
    sympy.stats.crv_types.GammaInverseDistribution: lambda a, b: {'a': a, 'scale': b},
    sympy.stats.crv_types.FDistributionDistribution: lambda d1, d2: {'dfn': d1, 'dfd': d2},
    sympy.stats.crv_types.LomaxDistribution: lambda alpha, lamb: {'c': alpha, 'scale': lamb},
    sympy.stats.crv_types.LogNormalDistribution: lambda mu, sigma: {'s': sigma, 'loc': mu},
    sympy.stats.crv_types.MaxwellDistribution: lambda a: {'loc': 0, 'scale': a},
    sympy.stats.crv_types.NakagamiDistribution: lambda mu, omega: {'nu': mu, 'scale': np.sqrt(omega)},
    sympy.stats.crv_types.TrapezoidalDistribution: lambda a, b, c, d: {
        'loc': a,
        'scale': d - a,
        'c': (b - a) / (d - a),
        'd': (c - a) / (d - a),
    },
    sympy.stats.crv_types.TriangularDistribution: lambda a, b, c: {'loc': a, 'scale': b - a, 'c': (c - a) / (b - a)},
    sympy.stats.crv_types.UniformDistribution: lambda a, b: {'loc': a, 'scale': b - a},
    sympy.stats.crv_types.GaussianInverseDistribution: lambda mu, lamb: {'mu': mu / lamb, 'scale': lamb},
    sympy.stats.crv_types.ExGaussianDistribution: lambda mean, std, rate: {
        'loc': mean,
        'scale': std,
        'K': 1 / (std * rate),
    },
}


def extract_distribution_parameters(
    parameters: dict[str, Any], distribution_args: Iterable[sympy.Symbol]
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Extract the distribution parameters from the other parameters."""
    distribution_parameters = {
        str(key): parameters.pop(str(key)) for key in distribution_args if str(key) in parameters
    }
    return distribution_parameters, parameters


def compute_integral_quad(
    integrand: sympy.Expr,
    lower_bound: float,
    upper_bound: float,
    true_positive_rate: float,
    false_positive_rate: float,
    random_var: sympy.Symbol,
) -> float:
    """Compute the integral using scipy quadrature for one stochastic variable."""
    integrand = integrand.subs('F_0', true_positive_rate).subs('F_1', false_positive_rate).evalf()
    if not integrand.free_symbols:  # if the integrand is constant, no need to call quad
        if integrand == 0:  # need this separate path since sometimes upper or lower bound can be infinite
            return 0
        return float(integrand * (upper_bound - lower_bound))
    integrand_fn = lambdify(random_var, integrand)
    result, _ = quad(integrand_fn, lower_bound, upper_bound)
    result: float
    return result


def compute_integral_multiple_quad(
    profit_integrand: sympy.Expr,
    rate_integrand: sympy.Expr,
    bounds: Iterable[float],
    true_positive_rates: Iterable[float],
    false_positive_rates: Iterable[float],
    random_variables: Iterable[sympy.Symbol],
    n_random: int,
) -> float:
    """Compute the integral using scipy quadrature for multiple stochastic variables."""
    profit_integrands = [
        lambdify(random_variables, profit_integrand.subs('F_0', tpr).subs('F_1', fpr).evalf())
        for tpr, fpr in zip(true_positive_rates, false_positive_rates, strict=True)
    ]

    if rate_integrand is None:  # compute maximum profit

        def integrand_fn(*random_vars: float) -> float:
            return float(max(integrand(*reversed(random_vars)) for integrand in profit_integrands))

    else:  # compute optimal rate
        rate_integrands = [
            lambdify(random_variables, rate_integrand.subs('F_0', tpr).subs('F_1', fpr).evalf())
            for tpr, fpr in zip(true_positive_rates, false_positive_rates, strict=True)
        ]

        def integrand_fn(*random_vars: float) -> float:
            best_index = np.argmax([integrand(*reversed(random_vars)) for integrand in profit_integrands])
            return float(rate_integrands[best_index](*reversed(random_vars)))

    if n_random == 1:
        result, _ = quad(integrand_fn, *bounds)
    elif n_random == 2:
        result, _ = dblquad(integrand_fn, *bounds)
    elif n_random == 3:
        result, _ = tplquad(integrand_fn, *bounds)
    else:
        result, _ = nquad(integrand_fn, *bounds)
    return float(result)


def _build_max_profit_score(
    tp_benefit: sympy.Expr,
    tn_benefit: sympy.Expr,
    fp_cost: sympy.Expr,
    fn_cost: sympy.Expr,
    integration_method: str,
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> MetricFn:
    random_symbols, deterministic_symbols = _identify_symbols(tp_benefit, tn_benefit, fp_cost, fn_cost)
    n_random = len(random_symbols)

    profit_function = _build_profit_function(
        tp_benefit=tp_benefit, tn_benefit=tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
    )
    if n_random == 0:
        max_profit_score = _build_max_profit_deterministic(profit_function, deterministic_symbols)
    else:
        max_profit_score = _build_max_profit_stochastic(
            profit_function,
            random_symbols,
            deterministic_symbols,
            integration_method=integration_method,
            n_mc_samples=n_mc_samples,
            rng=rng,
        )
    return max_profit_score


def _to_threshold_function(rate_function: RateFn) -> ThresholdFn:
    def threshold_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        rate = rate_function(y_true, y_score, **kwargs)
        return classification_threshold(y_true, y_score, rate)

    return threshold_function


def _build_max_profit_optimal_threshold(
    tp_benefit: sympy.Expr,
    tn_benefit: sympy.Expr,
    fp_cost: sympy.Expr,
    fn_cost: sympy.Expr,
    integration_method: str,
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> ThresholdFn:
    rate_fn = _build_max_profit_optimal_rate(
        tp_benefit=tp_benefit,
        tn_benefit=tn_benefit,
        fp_cost=fp_cost,
        fn_cost=fn_cost,
        integration_method=integration_method,
        n_mc_samples=n_mc_samples,
        rng=rng,
    )
    return _to_threshold_function(rate_fn)


def _build_max_profit_optimal_rate(
    tp_benefit: sympy.Expr,
    tn_benefit: sympy.Expr,
    fp_cost: sympy.Expr,
    fn_cost: sympy.Expr,
    integration_method: str,
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> RateFn:
    random_symbols, deterministic_symbols = _identify_symbols(tp_benefit, tn_benefit, fp_cost, fn_cost)
    n_random = len(random_symbols)

    profit_function = _build_profit_function(
        tp_benefit=tp_benefit, tn_benefit=tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
    )
    rate_function = _build_rate_function()
    if n_random == 0:
        optimal_rate = _build_max_profit_rate_deterministic(profit_function, deterministic_symbols)
    else:
        optimal_rate = _build_max_profit_rate_stochastic(
            profit_function, rate_function, random_symbols, deterministic_symbols, integration_method, n_mc_samples, rng
        )
    return optimal_rate


def _identify_symbols(
    tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> tuple[list[sympy.Symbol], list[sympy.Symbol]]:
    """Identify random and deterministic symbols in the profit function."""
    terms = tp_benefit + tn_benefit + fp_cost + fn_cost
    random_symbols = [symbol for symbol in terms.free_symbols if is_random(symbol)]
    deterministic_symbols = [symbol for symbol in terms.free_symbols if not is_random(symbol)]
    return random_symbols, deterministic_symbols


def _build_profit_function(
    tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> sympy.Expr:
    pos_prior, neg_prior, tpr, fpr = sympy.symbols('pi_0 pi_1 F_0 F_1')
    profit_function = (
        tp_benefit * pos_prior * tpr
        + tn_benefit * neg_prior * (1 - fpr)
        - fn_cost * pos_prior * (1 - tpr)
        - neg_prior * fp_cost * fpr
    )
    return profit_function


def _build_rate_function() -> sympy.Expr:
    pos_prior, neg_prior, tpr, fpr = sympy.symbols('pi_0 pi_1 F_0 F_1')
    return pos_prior * tpr + neg_prior * fpr


def _calculate_profits_deterministic(
    y_true: FloatNDArray, y_score: FloatNDArray, calculate_profit: Callable[..., float], **kwargs: Any
) -> tuple[FloatNDArray, FloatNDArray, FloatNDArray, float, float]:
    pi0 = float(np.mean(y_true))
    pi1 = 1 - pi0
    tprs, fprs = _compute_convex_hull(y_true, y_score)

    profits = np.zeros_like(tprs)
    for i, (tpr, fpr) in enumerate(zip(tprs, fprs, strict=False)):
        profits[i] = calculate_profit(pi_0=pi0, pi_1=pi1, F_0=tpr, F_1=fpr, **kwargs)

    return profits, tprs, fprs, pi0, pi1


def _build_max_profit_deterministic(
    profit_function: sympy.Expr, deterministic_symbols: Iterable[sympy.Symbol]
) -> MetricFn:
    """Compute the maximum profit for all deterministic variables."""
    calculate_profit = lambdify(list(profit_function.free_symbols), profit_function)

    @_check_parameters_decorator(*deterministic_symbols)
    def score_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        profits, *_ = _calculate_profits_deterministic(y_true, y_score, calculate_profit, **kwargs)
        return float(profits.max())

    return score_function


def _calculate_optimal_rate_deterministic(
    y_true: FloatNDArray, y_score: FloatNDArray, calculate_profit: Callable[..., float], **kwargs: Any
) -> float:
    profits, tprs, fprs, pi0, pi1 = _calculate_profits_deterministic(y_true, y_score, calculate_profit, **kwargs)
    best_index = np.argmax(profits)
    return float(tprs[best_index] * pi0 + fprs[best_index] * pi1)


def _build_max_profit_rate_deterministic(
    profit_function: sympy.Expr, deterministic_symbols: Iterable[sympy.Symbol]
) -> RateFn:
    """Calculate the optimal predicted positive rate."""
    calculate_profit = lambdify(list(profit_function.free_symbols), profit_function)

    @_check_parameters_decorator(*deterministic_symbols)
    def rate_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        return _calculate_optimal_rate_deterministic(y_true, y_score, calculate_profit, **kwargs)

    return rate_function


def _build_max_profit_rate_stochastic(
    profit_function: sympy.Expr,
    rate_function: sympy.Expr,
    random_symbols: Sequence[sympy.Symbol],
    deterministic_symbols: Iterable[sympy.Symbol],
    integration_method: str,
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> RateFn:
    """Compute the maximum profit for one or more stochastic variables."""
    n_random = len(random_symbols)
    if integration_method == 'auto':
        if n_random == 1:
            return _build_max_profit_rate_stochastic_piecewise(
                profit_function, rate_function, random_symbols[0], deterministic_symbols
            )
        elif n_random == 2:
            return _build_max_profit_stochastic_quad(
                profit_function, rate_function, random_symbols, deterministic_symbols
            )
        elif _support_all_distributions(random_symbols):
            return _build_max_profit_stochastic_qmc(
                profit_function, rate_function, random_symbols, deterministic_symbols, n_mc_samples, rng
            )
        else:
            return _build_max_profit_stochastic_mc(
                profit_function, rate_function, random_symbols, deterministic_symbols, n_mc_samples, rng
            )
    elif integration_method == 'quad':
        return _build_max_profit_stochastic_quad(profit_function, rate_function, random_symbols, deterministic_symbols)
    elif integration_method == 'monte-carlo':
        return _build_max_profit_stochastic_mc(
            profit_function, rate_function, random_symbols, deterministic_symbols, n_mc_samples, rng
        )
    elif integration_method == 'quasi-monte-carlo':
        return _build_max_profit_stochastic_qmc(
            profit_function, rate_function, random_symbols, deterministic_symbols, n_mc_samples, rng
        )
    else:
        raise ValueError(f'Integration method {integration_method} is not supported')


def compute_piecewise_bounds(
    compute_bounds: Callable[..., float],
    true_positive_rates: FloatNDArray,
    false_positive_rates: FloatNDArray,
    positive_class_prior: float,
    negative_class_prior: float,
    random_var_bounds: tuple[float | sympy.Expr, ...],
    distribution_parameters: dict[str, Any],
    **kwargs: Any,
) -> tuple[list[float], float, float]:
    """
    Compute the consecutive bounds of the stochastic variable for which the expected profit is equal.

    These bounds can then be used during piecewise integration.
    """
    bounds = []
    for (tpr0, fpr0), (tpr1, fpr1) in islice(
        pairwise(zip(true_positive_rates, false_positive_rates, strict=False)), len(true_positive_rates) - 2
    ):
        bounds.append(
            compute_bounds(
                F_0=tpr0,
                F_1=fpr0,
                F_2=tpr1,
                F_3=fpr1,
                pi_0=positive_class_prior,
                pi_1=negative_class_prior,
                **kwargs,
            )
        )

    # bounds of the random variable can be parameterized by the user
    # if so substitute the parameters in the bounds with the user-provided values
    if isinstance(upper_bound := random_var_bounds[1], sympy.Expr):
        upper_bound = upper_bound.subs(distribution_parameters)
    bounds.append(upper_bound)
    if isinstance(lower_bound := random_var_bounds[0], sympy.Expr):
        lower_bound = lower_bound.subs(distribution_parameters)
    bounds.insert(0, lower_bound)
    return bounds, upper_bound, lower_bound


def _build_max_profit_rate_stochastic_piecewise(
    profit_function: sympy.Expr,
    rate_function: sympy.Expr,
    random_symbol: sympy.Symbol,
    deterministic_symbols: Iterable[sympy.Symbol],
) -> RateFn:
    """
    Compute the optimal predicted positive rate for a single stochastic variable using piecewise integration.

    For each convex hull segment, the bounds of the random variable are computed
    in which that decision threshold is optimal.
    For each segment, the profit is integrated over the bounds of the random variable.
    """
    profit_prime = profit_function.subs('F_0', 'F_2').subs('F_1', 'F_3')
    bound_eq = solve(profit_function - profit_prime, random_symbol)[0]
    compute_bounds = lambdify(list(bound_eq.free_symbols), bound_eq)

    random_var_bounds = pspace(random_symbol).domain.set.args
    distribution_args = pspace(random_symbol).distribution.args

    integrand = rate_function * density(random_symbol).pdf(random_symbol)

    if all(isinstance(arg, sympy.core.numbers.Integer) for arg in distribution_args):
        dist_params = []
    else:
        dist_params = [arg for arg in distribution_args if not isinstance(arg, sympy.core.numbers.Integer)]

    @_check_parameters_decorator(*deterministic_symbols, *dist_params)
    def calculate_rate_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        positive_class_prior = float(np.mean(y_true))
        negative_class_prior = 1 - positive_class_prior
        true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)

        # distribution parameters of the random variable
        distribution_parameters, kwargs = extract_distribution_parameters(kwargs, distribution_args)
        bounds, upper_bound, lower_bound = compute_piecewise_bounds(
            compute_bounds,
            true_positive_rates,
            false_positive_rates,
            positive_class_prior,
            negative_class_prior,
            random_var_bounds,
            distribution_parameters,
            **kwargs,
        )

        integrand_ = (
            integrand.subs(distribution_parameters)
            .subs('pi_0', positive_class_prior)
            .subs('pi_1', negative_class_prior)
        )
        score = 0.0
        for (lower_bound, upper_bound), tpr, fpr in zip(
            pairwise(bounds), true_positive_rates, false_positive_rates, strict=False
        ):
            score += compute_integral_quad(integrand_, lower_bound, upper_bound, tpr, fpr, random_symbol)
        return score

    return calculate_rate_function


def _support_all_distributions(random_symbols: Iterable[sympy.Symbol]) -> bool:
    return all(pspace(r).distribution.__class__ in _sympy_dist_to_scipy for r in random_symbols)


def _build_max_profit_stochastic(
    profit_function: sympy.Expr,
    random_symbols: Sequence[sympy.Symbol],
    deterministic_symbols: Iterable[sympy.Symbol],
    integration_method: str,
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> MetricFn:
    """Compute the maximum profit for one or more stochastic variables."""
    n_random = len(random_symbols)
    if integration_method == 'auto':
        if n_random == 1:
            return _build_max_profit_stochastic_piecewise(profit_function, random_symbols[0], deterministic_symbols)
        elif n_random == 2:
            return _build_max_profit_stochastic_quad(profit_function, None, random_symbols, deterministic_symbols)
        elif _support_all_distributions(random_symbols):
            return _build_max_profit_stochastic_qmc(
                profit_function, None, random_symbols, deterministic_symbols, n_mc_samples, rng
            )
        else:
            return _build_max_profit_stochastic_mc(
                profit_function, None, random_symbols, deterministic_symbols, n_mc_samples, rng
            )
    elif integration_method == 'quad':
        return _build_max_profit_stochastic_quad(profit_function, None, random_symbols, deterministic_symbols)
    elif integration_method == 'monte-carlo':
        return _build_max_profit_stochastic_mc(
            profit_function, None, random_symbols, deterministic_symbols, n_mc_samples, rng
        )
    elif integration_method == 'quasi-monte-carlo':
        return _build_max_profit_stochastic_qmc(
            profit_function, None, random_symbols, deterministic_symbols, n_mc_samples, rng
        )
    else:
        raise ValueError(f'Integration method {integration_method} is not supported')


def _build_max_profit_stochastic_piecewise(
    profit_function: sympy.Expr, random_symbol: sympy.Symbol, deterministic_symbols: Iterable[sympy.Symbol]
) -> MetricFn:
    """
    Compute the maximum profit for a single stochastic variable using piecewise integration.

    For each convex hull segment, the bounds of the random variable are computed
    in which that decision threshold is optimal.
    For each segment, the profit is integrated over the bounds of the random variable.
    """
    profit_prime = profit_function.subs('F_0', 'F_2').subs('F_1', 'F_3')
    bound_eq = solve(profit_function - profit_prime, random_symbol)[0]
    compute_bounds = lambdify(list(bound_eq.free_symbols), bound_eq)

    random_var_bounds = pspace(random_symbol).domain.set.args
    distribution_args = pspace(random_symbol).distribution.args

    integrand = profit_function * density(random_symbol).pdf(random_symbol)

    if all(isinstance(arg, sympy.core.numbers.Integer) for arg in distribution_args):
        dist_params = []
    else:
        dist_params = [arg for arg in distribution_args if not isinstance(arg, sympy.core.numbers.Integer)]

    @_check_parameters_decorator(*deterministic_symbols, *dist_params)
    def score_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        positive_class_prior = float(np.mean(y_true))
        negative_class_prior = 1 - positive_class_prior
        true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)

        # distribution parameters of the random variable
        distribution_parameters, kwargs = extract_distribution_parameters(kwargs, distribution_args)
        bounds, upper_bound, lower_bound = compute_piecewise_bounds(
            compute_bounds,
            true_positive_rates,
            false_positive_rates,
            positive_class_prior,
            negative_class_prior,
            random_var_bounds,
            distribution_parameters,
            **kwargs,
        )

        integrand_ = (
            integrand.subs(kwargs)
            .subs(distribution_parameters)
            .subs('pi_0', positive_class_prior)
            .subs('pi_1', negative_class_prior)
        )
        score = 0.0
        for (lower_bound, upper_bound), tpr, fpr in zip(
            pairwise(bounds), true_positive_rates, false_positive_rates, strict=False
        ):
            score += compute_integral_quad(integrand_, lower_bound, upper_bound, tpr, fpr, random_symbol)
        return score

    return score_function


def _build_max_profit_stochastic_quad(
    profit_function: sympy.Expr,
    rate_function: sympy.Expr | None,
    random_symbols: Sequence[sympy.Symbol],
    deterministic_symbols: Iterable[sympy.Symbol],
) -> RateFn | MetricFn:
    """
    Compute the optimal predicted positive rate for one or more stochastic variables using quad integration.

    This method is very slow for more than 2 stochastic variables.
    It is recommended to use Quasi Monte Carlo integration for more than 2 stochastic variables.
    """
    n_random = len(random_symbols)
    random_variables_bounds = [pspace(random_symbol).domain.set.args for random_symbol in random_symbols]
    random_variables_bounds = [(lb, up) for (lb, up, *_) in random_variables_bounds]
    distributions_args = [pspace(random_symbol).distribution.args for random_symbol in random_symbols]
    distribution_args = [arg for args in distributions_args for arg in args]

    for random_symbol in random_symbols:
        profit_function *= density(random_symbol).pdf(random_symbol)
    if rate_function is not None:
        for random_symbol in random_symbols:
            rate_function *= density(random_symbol).pdf(random_symbol)

    if all(isinstance(arg, sympy.core.numbers.Integer) for arg in distribution_args):
        dist_params = []
    else:
        dist_params = [arg for arg in distribution_args if not isinstance(arg, sympy.core.numbers.Integer)]

    @_check_parameters_decorator(*deterministic_symbols, *dist_params)
    def calculate_integral(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        positive_class_prior = float(np.mean(y_true))
        negative_class_prior = 1 - positive_class_prior
        true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)

        # certain distributions determine the bounds of the integral (e.g., uniform)
        # for those distributions we have to fill in the parameters of the distribution
        distribution_parameters, kwargs = extract_distribution_parameters(kwargs, distribution_args)
        bounds = [bound for bounds in random_variables_bounds for bound in bounds]
        bounds = [
            bounds.subs(distribution_parameters) if isinstance(bounds, sympy.Expr) else bounds for bounds in bounds
        ]

        profit_integrand_ = (
            profit_function.subs(kwargs)
            .subs(distribution_parameters)
            .subs('pi_0', positive_class_prior)
            .subs('pi_1', negative_class_prior)
        )
        if rate_function is not None:
            rate_integrand_ = (
                rate_function.subs(kwargs)
                .subs(distribution_parameters)
                .subs('pi_0', positive_class_prior)
                .subs('pi_1', negative_class_prior)
            )
        else:
            rate_integrand_ = None
        return compute_integral_multiple_quad(
            profit_integrand_,
            rate_integrand_,
            bounds,
            true_positive_rates,
            false_positive_rates,
            random_symbols,
            n_random,
        )

    return calculate_integral


def _build_max_profit_stochastic_mc(
    profit_function: sympy.Expr,
    rate_function: sympy.Expr | None,
    random_symbols: Sequence[sympy.Symbol],
    deterministic_symbols: Iterable[sympy.Symbol],
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> MetricFn | RateFn:
    """
    Compute the maximum profit for one or more stochastic variables using Monte Carlo (MC) integration.

    This method is less accurate than quad integration but faster for many stochastic variables.
    The QMC method is preferred over the MC due to better accuracy.
    This method should only be used if there is no mapping of sympy distributions to scipy distributions.
    """
    distributions_args = [pspace(random_symbol).distribution.args for random_symbol in random_symbols]
    distribution_args = [arg for args in distributions_args for arg in args]
    if all(isinstance(arg, sympy.core.numbers.Integer) for arg in distribution_args):
        param_grid_needs_recompute = False
        param_grid = [sympy.stats.sample(random_var, size=(n_mc_samples,), seed=rng) for random_var in random_symbols]
        dist_params = []
    else:
        cached_dist_params = {str(arg): arg for arg in distribution_args}
        param_grid_needs_recompute = True
        param_grid = None
        dist_params = [arg for arg in distribution_args if not isinstance(arg, sympy.core.numbers.Integer)]

    @_check_parameters_decorator(*deterministic_symbols, *dist_params)
    def score_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        positive_class_prior = float(np.mean(y_true))
        negative_class_prior = 1 - positive_class_prior
        true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)

        nonlocal param_grid
        nonlocal param_grid_needs_recompute
        nonlocal cached_dist_params
        if param_grid_needs_recompute:
            # distribution parameters of the random variable
            distribution_parameters, kwargs = extract_distribution_parameters(kwargs, distribution_args)
            if cached_dist_params != distribution_parameters:
                cached_dist_params = distribution_parameters
                param_grid = [
                    sympy.stats.sample(random_var.subs(cached_dist_params), size=(n_mc_samples,), seed=rng)
                    for random_var in random_symbols
                ]

            profit_integrand = (
                profit_function.subs(kwargs)
                .subs(cached_dist_params)
                .subs('pi_0', positive_class_prior)
                .subs('pi_1', negative_class_prior)
            )
            if rate_function is not None:
                rate_integrand = (
                    rate_function.subs(cached_dist_params)
                    .subs('pi_0', positive_class_prior)
                    .subs('pi_1', negative_class_prior)
                )
        else:
            profit_integrand = (
                profit_function.subs(kwargs).subs('pi_0', positive_class_prior).subs('pi_1', negative_class_prior)
            )
            if rate_function is not None:
                rate_integrand = rate_function.subs('pi_0', positive_class_prior).subs('pi_1', negative_class_prior)
        profit_integrands = [
            lambdify(random_symbols, profit_integrand.subs('F_0', tpr).subs('F_1', fpr).evalf())
            for tpr, fpr in zip(true_positive_rates, false_positive_rates, strict=True)
        ]

        results = np.empty((len(profit_integrands), n_mc_samples))
        for i, integrand in enumerate(profit_integrands):
            results[i, :] = integrand(*param_grid)
        if rate_function is None:
            result = results.max(axis=0).mean()
        else:
            rate_integrands = [
                lambdify(random_symbols, rate_integrand.subs('F_0', tpr).subs('F_1', fpr).evalf())
                for tpr, fpr in zip(true_positive_rates, false_positive_rates, strict=True)
            ]
            rate_results = np.empty((len(profit_integrands), n_mc_samples))
            best_indices = results.argmax(axis=0)
            for i, integrand in enumerate(rate_integrands):
                rate_results[i, :] = integrand(*param_grid)
            result = float(rate_results[best_indices, np.arange(n_mc_samples)].mean())

        return float(result)

    return score_function


def _build_max_profit_stochastic_qmc(
    profit_function: sympy.Expr,
    rate_function: sympy.Expr | None,
    random_symbols: Sequence[sympy.Symbol],
    deterministic_symbols: Iterable[sympy.Symbol],
    n_mc_samples: int,
    rng: np.random.RandomState,
) -> MetricFn | RateFn:
    distributions_args = [pspace(random_symbol).distribution.args for random_symbol in random_symbols]
    distribution_args = [arg for args in distributions_args for arg in args]
    # Generate a Sobol sequence for QMC sampling
    sobol = Sobol(d=len(random_symbols), scramble=True, seed=rng)
    sobol_samples = sobol.random(n_mc_samples)
    if all(isinstance(arg, sympy.core.numbers.Integer) for arg in distribution_args):
        # If all distribution parameters are fixed, then the param grid can be pre-computed.
        param_grid_needs_recompute = False
        # convert to scipy distributions
        scipy_distributions = []
        for random_var in random_symbols:
            sympy_distribution = pspace(random_var).distribution.__class__
            scipy_distribution = _sympy_dist_to_scipy[sympy_distribution]
            sympy_dist_params = [float(arg) for arg in pspace(random_var).distribution.args]
            if sympy_distribution in _sympy_dist_to_scipy_params:
                scipy_dist_kwargs = _sympy_dist_to_scipy_params[sympy_distribution](*sympy_dist_params)
                scipy_distributions.append(scipy_distribution(**scipy_dist_kwargs))
            else:
                scipy_dist_params = sympy_dist_params
                scipy_distributions.append(scipy_distribution(*scipy_dist_params))

        # scipy_distributions = [
        #     _sympy_dist_to_scipy[pspace(random_var).distribution.__class__](*[
        #         float(arg) for arg in pspace(random_var).distribution.args
        #     ])
        #     for random_var in random_symbols
        # ]
        param_grid = [dist.ppf(sobol_samples[:, i]) for i, dist in enumerate(scipy_distributions)]
        dist_params = []

    else:
        cached_dist_params = {str(arg): arg for arg in distribution_args}
        scipy_distributions = None
        param_grid_needs_recompute = True
        param_grid = []
        dist_params = [arg for arg in distribution_args if not isinstance(arg, sympy.core.numbers.Integer)]

    @_check_parameters_decorator(*deterministic_symbols, *dist_params)
    def score_function(y_true: FloatNDArray, y_score: FloatNDArray, **kwargs: Any) -> float:
        positive_class_prior = float(np.mean(y_true))
        negative_class_prior = 1 - positive_class_prior
        true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)

        nonlocal param_grid
        nonlocal param_grid_needs_recompute
        nonlocal cached_dist_params
        nonlocal scipy_distributions
        if param_grid_needs_recompute:
            # distribution parameters of the random variable
            distribution_parameters, kwargs = extract_distribution_parameters(kwargs, distribution_args)
            if cached_dist_params != distribution_parameters:
                cached_dist_params = distribution_parameters
                scipy_distributions = []
                for random_var in random_symbols:
                    sympy_distribution = pspace(random_var).distribution.__class__
                    scipy_distribution = _sympy_dist_to_scipy[sympy_distribution]
                    sympy_dist_params = [
                        float(arg) for arg in pspace(random_var.subs(cached_dist_params)).distribution.args
                    ]
                    if sympy_distribution in _sympy_dist_to_scipy_params:
                        scipy_dist_kwargs = _sympy_dist_to_scipy_params[sympy_distribution](*sympy_dist_params)
                        scipy_distributions.append(scipy_distribution(**scipy_dist_kwargs))
                    else:
                        scipy_dist_params = sympy_dist_params
                        scipy_distributions.append(scipy_distribution(*scipy_dist_params))
                # scipy_distributions = [
                #     _sympy_dist_to_scipy[pspace(random_var).distribution.__class__](*[
                #         float(arg) for arg in pspace(random_var.subs(cached_dist_params)).distribution.args
                #     ])
                #     for random_var in random_symbols
                # ]
                param_grid = [dist.ppf(sobol_samples[:, i]) for i, dist in enumerate(scipy_distributions)]

            profit_integrand = (
                profit_function.subs(kwargs)
                .subs(cached_dist_params)
                .subs('pi_0', positive_class_prior)
                .subs('pi_1', negative_class_prior)
            )
            if rate_function is not None:
                rate_integrand = (
                    rate_function.subs(cached_dist_params)
                    .subs('pi_0', positive_class_prior)
                    .subs('pi_1', negative_class_prior)
                )
        else:
            profit_integrand = (
                profit_function.subs(kwargs).subs('pi_0', positive_class_prior).subs('pi_1', negative_class_prior)
            )
            if rate_function is not None:
                rate_integrand = rate_function.subs('pi_0', positive_class_prior).subs('pi_1', negative_class_prior)

        profit_integrands = [
            lambdify(random_symbols, profit_integrand.subs('F_0', tpr).subs('F_1', fpr).evalf())
            for tpr, fpr in zip(true_positive_rates, false_positive_rates, strict=True)
        ]

        results = np.empty((len(profit_integrands), n_mc_samples))
        for i, integrand in enumerate(profit_integrands):
            results[i, :] = integrand(*param_grid)
        if rate_function is None:
            result = float(results.max(axis=0).mean())
        else:
            rate_integrands = [
                lambdify(random_symbols, rate_integrand.subs('F_0', tpr).subs('F_1', fpr).evalf())
                for tpr, fpr in zip(true_positive_rates, false_positive_rates, strict=True)
            ]
            rate_results = np.empty((len(profit_integrands), n_mc_samples))
            best_indices = results.argmax(axis=0)
            for i, integrand in enumerate(rate_integrands):
                rate_results[i, :] = integrand(*param_grid)
            result = float(rate_results[best_indices, np.arange(n_mc_samples)].mean())

        return result

    return score_function


def _max_profit_score_to_latex(
    tp_benefit: sympy.Expr, tn_benefit: sympy.Expr, fp_cost: sympy.Expr, fn_cost: sympy.Expr
) -> str:
    from sympy.printing.latex import latex

    profit_function = _build_profit_function(
        tp_benefit=-tp_benefit, tn_benefit=-tn_benefit, fp_cost=fp_cost, fn_cost=fn_cost
    )
    random_symbols = [symbol for symbol in profit_function.free_symbols if is_random(symbol)]

    if random_symbols:
        integrand = profit_function
        for random_symbol in random_symbols:
            integrand *= density(random_symbol).pdf(random_symbol)
        integral = integrand
        for random_symbol in random_symbols:
            lower_bound, upper_bound = pspace(random_symbol).domain.set.args[:2]
            integral = sympy.Integral(integral, (random_symbol, lower_bound, upper_bound))

        output = latex(integral, mode='plain', order=None)
    else:
        output = latex(profit_function, mode='plain', order=None)
    return f'$\\displaystyle {output}$'



from ..._types import FloatArrayLike, FloatNDArray
from .._validation import _check_fraction, _check_positive, _check_shape, _check_y_pred, _check_y_true


def _validate_input(y_true: FloatArrayLike, y_pred: FloatArrayLike, roi: float) -> tuple[FloatNDArray, FloatNDArray]:
    y_true = _check_y_true(y_true)
    y_pred = _check_y_pred(y_pred)
    _check_shape(y_true, y_pred)
    _check_positive(roi, 'roi')
    return y_true, y_pred


def _validate_input_mp(
    y_true: FloatArrayLike, y_pred: FloatArrayLike, default_prob: float, roi: float
) -> tuple[FloatNDArray, FloatNDArray]:
    _check_fraction(default_prob, 'default_prob')
    return _validate_input(y_true, y_pred, roi)


def _validate_input_emp(
    y_true: FloatArrayLike, y_pred: FloatArrayLike, p_0: float, p_1: float, roi: float
) -> tuple[FloatNDArray, FloatNDArray]:
    _check_fraction(p_0, 'p_0')
    _check_fraction(p_1, 'p_1')
    return _validate_input(y_true, y_pred, roi)



import numpy as np
from scipy.spatial import ConvexHull, QhullError
from sklearn.metrics import roc_curve

from .._types import FloatNDArray


def _compute_convex_hull(
    y_true: FloatNDArray, y_pred: FloatNDArray, expand_dims: bool = False
) -> tuple[FloatNDArray, FloatNDArray]:
    """
    Compute the convex hull points of the ROC curve.

    Parameters
    ----------
    y_true : 1D np.ndarray, shape=(n_samples,)
        Binary target values.

    y_pred : 1D np.ndarray, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    expand_dims : bool, default=False
        Whether to expand the dimensions of the convex hull points to (n_points, 1).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Convex Hull points of the ROC curve (TPR, FPR)
    """
    fpr, tpr, _ = roc_curve(y_true, y_pred, pos_label=1, drop_intermediate=True)
    if fpr[0] != 0 or tpr[0] != 0:
        fpr = np.concatenate([[0], fpr])
        tpr = np.concatenate([[0], tpr])
    if fpr[-1] != 1 or tpr[-1] != 1:
        fpr = np.concatenate([fpr, [1]])
        tpr = np.concatenate([tpr, [1]])

    is_finite = np.isfinite(fpr) & np.isfinite(tpr)
    fpr = fpr[is_finite]
    tpr = tpr[is_finite]
    if fpr.shape[0] < 2:
        raise ValueError('Too few distinct predictions for ROCCH')

    points = np.c_[fpr, tpr]  # concatenate into matrix with two columns
    try:
        ind = ConvexHull(points).vertices  # indices of the points on the convex hull
    except QhullError:
        return np.array([0, 1]), np.array([0, 1])

    convex_hull_fpr = fpr[ind]
    convex_hull_tpr = tpr[ind]
    ind_upper_triangle = convex_hull_fpr < convex_hull_tpr  # only consider points above the 45° line
    convex_hull_fpr = np.concatenate([[0], convex_hull_fpr[ind_upper_triangle], [1]])
    convex_hull_tpr = np.concatenate([[0], convex_hull_tpr[ind_upper_triangle], [1]])
    ind = np.argsort(convex_hull_fpr)  # sort along the x-axis
    convex_hull_fpr = convex_hull_fpr[ind]
    convex_hull_tpr = convex_hull_tpr[ind]

    if expand_dims:
        convex_hull_tpr = np.expand_dims(convex_hull_tpr, axis=1)
        convex_hull_fpr = np.expand_dims(convex_hull_fpr, axis=1)

    return convex_hull_tpr, convex_hull_fpr



import numpy as np

from .._types import FloatArrayLike, FloatNDArray
from ._validation import _check_binary, _check_fraction, _check_shape, _check_variance, _check_y_pred, _check_y_true


def _validate_input(
    y_true: FloatArrayLike, y_score: FloatArrayLike, fraction: float
) -> tuple[FloatNDArray, FloatNDArray]:
    y_true: FloatNDArray = _check_y_true(y_true)
    y_score: FloatNDArray = _check_y_pred(y_score)
    _check_shape(y_true, y_score)
    _check_binary(y_true)
    _check_variance(y_true)
    _check_shape(y_true, y_score)
    _check_fraction(fraction, 'fraction')

    return y_true, y_score


def lift_score(
    y_true: FloatArrayLike, y_score: FloatArrayLike, *, fraction: float = 0.1, check_input: bool = True
) -> float:
    """
    Compute the lift score for the top fraction of predictions.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('positive': 1, 'negative': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    fraction : float, optional, default: 0.1
        Fraction of data to consider. Must be between 0 and 1.

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    lift_score : float
        Lift score for the top fraction of the data.
    """
    if check_input:
        y_true, y_score = _validate_input(y_true, y_score, fraction)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)

    # Sort the predictions in descending order
    sorted_indices = np.argsort(y_score)[::-1]
    sorted_labels = y_true[sorted_indices]
    top_fraction = int(round(len(sorted_labels) * fraction, 0))

    n_positives_top_fraction: int = np.sum(sorted_labels[:top_fraction])
    prop_positives_top_fraction = n_positives_top_fraction / top_fraction

    return prop_positives_top_fraction / float(np.mean(y_true))



import sys
import warnings
from collections.abc import Callable, Sequence
from functools import partial
from numbers import Real
from typing import Any, ClassVar, Literal, TypeVar, overload

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.special import expit
from sklearn.base import clone
from sklearn.utils._param_validation import HasMethods

from ..._types import FloatArrayLike, FloatNDArray, ParameterConstraint
from ...metrics.metric.prebuilt_metrics import make_generic_cost_metric

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

try:
    from xgboost import XGBClassifier
except ImportError:
    XGBClassifier = TypeVar('XGBClassifier')  # type: ignore[misc, assignment]
try:
    from lightgbm import LGBMClassifier
except ImportError:
    LGBMClassifier = TypeVar('LGBMClassifier')  # type: ignore[misc, assignment]
try:
    from catboost import CatBoostClassifier
except ImportError:
    CatBoostClassifier = TypeVar('CatBoostClassifier')  # type: ignore[misc, assignment]

from ..._common import Parameter
from ...metrics import Metric
from ...metrics._loss import cy_boost_grad_hess
from .._base import BaseBoostClassifier
from ._cs_mixin import CostSensitiveMixin

# Hessian is 0 at score 0.5
# which means that at initialization the model optimization doesn't do anything
# therefore we add a small nudge which kickstarts the optimization algorithm (so hessian is not 0)
_BASE_SCORE = 0.5 + 1e-2


class LGBMObjective:
    """AEC objective for lightgbm."""

    def __init__(self, gradient_const: FloatNDArray):
        self.gradient_const = gradient_const

    def __call__(self, y_true: FloatNDArray, y_score: FloatNDArray) -> tuple[FloatNDArray, FloatNDArray]:
        """
        Create an objective function for the AEC measure.

        Parameters
        ----------
        y_true : np.ndarray
            Ground truth labels
        y_score : np.ndarray
            Predicted labels

        Returns
        -------
        gradient  : np.ndarray
            Gradient of the objective function.

        hessian : np.ndarray
            Hessian of the objective function.
        """
        gradient: FloatNDArray
        hessian: FloatNDArray
        gradient, hessian = cy_boost_grad_hess(y_true, y_score, self.gradient_const)
        return gradient, hessian


class CSBoostClassifier(BaseBoostClassifier, CostSensitiveMixin):
    """
    Cost-sensitive gradient boosting classifier.

    CSBoostClassifier supports :class:`xgboost:xgboost.XGBClassifier`, :class:`lightgbm:lightgbm.LGBMClassifier`
    and :class:`catboost.CatBoostClassifier` as base estimators.
    By default, it uses XGBoost classifier with default hyperparameters.

    Read more in the :ref:`User Guide <csboost>`.

    .. seealso::

        :func:`~empulse.metrics.make_objective_aec` : Creates the instance-dependent cost function.

        :class:`~empulse.models.CSLogitClassifier` : Cost-sensitive logistic regression.

    Parameters
    ----------
    estimator : :class:`xgboost:xgboost.XGBClassifier`, :class:`lightgbm:lightgbm.LGBMClassifier` \
    or :class:`catboost.CatBoostClassifier`, optional
        XGBoost or LightGBM classifier to be fit with desired hyperparameters.
        If not provided, a XGBoost classifier with default hyperparameters is used.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    loss : :class:`empulse.metrics.Metric`, default=None
        Loss function to optimize. Metric parameters are passed as ``loss_params``
          to the :Meth:`~empulse.models.CSBoostClassifier.fit` method.

    Attributes
    ----------
    classes_ : numpy.ndarray, shape=(n_classes,)
        Unique classes in the target.

    estimator_ : :class:`xgboost:xgboost.XGBClassifier`
        Fitted XGBoost classifier.

    Examples
    --------

    .. code-block:: python

        import numpy as np
        from empulse.models import CSBoostClassifier
        from sklearn.datasets import make_classification

        X, y = make_classification()
        fn_cost = np.random.rand(y.size)  # instance-dependent cost
        fp_cost = 5  # constant cost

        model = CSBoostClassifier()
        model.fit(X, y, fn_cost=fn_cost, fp_cost=fp_cost)
        y_proba = model.predict_proba(X)

    Example with passing instance-dependent costs through cross-validation:

    .. code-block:: python

        import numpy as np
        from empulse.models import CSBoostClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import cross_val_score
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler

        set_config(enable_metadata_routing=True)

        X, y = make_classification()
        fn_cost = np.random.rand(y.size)
        fp_cost = 5

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('model', CSBoostClassifier().set_fit_request(fn_cost=True, fp_cost=True))
        ])

        cross_val_score(pipeline, X, y, params={'fn_cost': fn_cost, 'fp_cost': fp_cost})

    Example with passing instance-dependent costs through a grid search:

    .. code-block:: python

        import numpy as np
        from empulse.metrics import expected_cost_loss
        from empulse.models import CSBoostClassifier
        from sklearn import set_config
        from sklearn.datasets import make_classification
        from sklearn.model_selection import GridSearchCV
        from sklearn.metrics import make_scorer
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from xgboost import XGBClassifier

        set_config(enable_metadata_routing=True)

        X, y = make_classification(n_samples=50)
        fn_cost = np.random.rand(y.size)
        fp_cost = 5

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('model', CSBoostClassifier(
                XGBClassifier(n_jobs=2, n_estimators=10)
            ).set_fit_request(fn_cost=True, fp_cost=True))
        ])
        param_grid = {
            'model__estimator__learning_rate': np.logspace(-5, 0, 5),
        }
        scorer = make_scorer(
            expected_cost_loss,
            response_method='predict_proba',
            greater_is_better=False,
            normalize=True
        )
        scorer = scorer.set_score_request(fn_cost=True, fp_cost=True)

        grid_search = GridSearchCV(pipeline, param_grid=param_grid, scoring=scorer)
        grid_search.fit(X, y, fn_cost=fn_cost, fp_cost=fp_cost)

    References
    ----------
    .. [1] Höppner, S., Baesens, B., Verbeke, W., & Verdonck, T. (2022).
           Instance-dependent cost-sensitive learning for detecting transfer fraud.
           European Journal of Operational Research, 297(1), 291-300.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        **BaseBoostClassifier._parameter_constraints,
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
        'loss': [HasMethods('_gradient_boost_objective'), None],
    }

    def __init__(
        self,
        estimator: XGBClassifier | LGBMClassifier | CatBoostClassifier | None = None,
        *,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
        loss: Metric | None = None,
    ) -> None:
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost
        self.loss = loss
        super().__init__(estimator=estimator)

    def fit(
        self,
        X: FloatArrayLike,
        y: ArrayLike,
        *,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fit_params: dict[str, Any] | None = None,
        **loss_params: Any,
    ) -> Self:
        """
        Fit the model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)

        y : array-like of shape (n_samples,)

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        fit_params : dict
            Additional keyword arguments to pass to the estimator's fit method.

        loss_params : dict
            Additional keyword arguments to pass to the loss function if using a custom loss function.

        Returns
        -------
        self : CSBoostClassifier
            Fitted CSBoost model.
        """
        super().fit(
            X,
            y,
            tp_cost=tp_cost,
            tn_cost=tn_cost,
            fn_cost=fn_cost,
            fp_cost=fp_cost,
            fit_params=fit_params,
            **loss_params,
        )
        return self

    def _fit(
        self,
        X: NDArray[Any],
        y: NDArray[Any],
        *,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
        fit_params: dict[str, Any] | None = None,
        **loss_params: Any,
    ) -> Self:
        if fit_params is None:
            fit_params = {}
        # allow sample weights still to be passed as kwargs to comply with sklearn interface
        if 'sample_weight' in loss_params:
            fit_params['sample_weight'] = loss_params.pop('sample_weight')

        if self.loss is None:
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost
            )

        if self.estimator is None:
            self._initialize_default_estimator(
                y=y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params
            )
        else:
            self._initialize_custom_estimator(
                y=y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params
            )

        if isinstance(self.estimator_, XGBClassifier):
            self.estimator_.fit(X, y, **fit_params)
        elif isinstance(self.estimator_, LGBMClassifier):
            self.estimator_.fit(X, y, init_score=np.full(y.shape, _BASE_SCORE), **fit_params)
        else:
            indices = np.arange(X.shape[0])
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    'ignore',
                    message='Can\'t optimize method "calc_ders_range" because self argument is used',
                    category=UserWarning,
                )
                warnings.filterwarnings(
                    'ignore',
                    message='Can\'t optimize method "evaluate" because self argument is used',
                    category=UserWarning,
                )
                if 'sample_weight' in fit_params:
                    raise ValueError('Sample weights are not allowed when training CatboostClassifier.')
                self.estimator_.fit(X, y, sample_weight=indices, baseline=np.full(y.shape, _BASE_SCORE), **fit_params)
        return self

    def _initialize_default_estimator(
        self,
        y: FloatNDArray,
        tp_cost: FloatNDArray | FloatArrayLike | float,
        tn_cost: FloatNDArray | FloatArrayLike | float,
        fn_cost: FloatNDArray | FloatArrayLike | float,
        fp_cost: FloatNDArray | FloatArrayLike | float,
        **loss_params: Any,
    ) -> None:
        if XGBClassifier is None:
            raise ImportError(
                f'XGBoost package is required to use {type(self).__name__}. '
                'Install optional dependencies through `pip install empulse[optional]` or '
                '`pip install xgboost`'
            )
        objective = self._get_objective(
            'xgboost', y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params
        )
        self.estimator_ = XGBClassifier(objective=objective, base_score=_BASE_SCORE)

    def _initialize_custom_estimator(
        self,
        y: FloatNDArray,
        tp_cost: FloatNDArray | FloatArrayLike | float,
        tn_cost: FloatNDArray | FloatArrayLike | float,
        fn_cost: FloatNDArray | FloatArrayLike | float,
        fp_cost: FloatNDArray | FloatArrayLike | float,
        **loss_params: Any,
    ) -> None:
        if isinstance(self.estimator, XGBClassifier):
            objective = self._get_objective(
                'xgboost', y=y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params
            )
            self.estimator_ = clone(self.estimator).set_params(objective=objective, base_score=_BASE_SCORE)
        elif isinstance(self.estimator, LGBMClassifier):
            objective = self._get_objective(
                'lightgbm', y=y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params
            )
            self.estimator_ = clone(self.estimator).set_params(objective=objective)
        elif isinstance(self.estimator, CatBoostClassifier):
            # self._initialize_catboost_estimator(tp_cost, tn_cost, fn_cost, fp_cost, **loss_params)
            loss_function, eval_metric = self._get_objective(
                'catboost', y=y, tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, **loss_params
            )
            self.estimator_ = clone(self.estimator).set_params(loss_function=loss_function, eval_metric=eval_metric)
        else:
            raise ValueError('Estimator must be an instance of XGBClassifier, LGBMClassifier, or CatBoostClassifier')

    @overload
    def _get_objective(
        self,
        framework: Literal['xgboost'],
        y: FloatNDArray,
        tp_cost: FloatNDArray | FloatArrayLike | float,
        tn_cost: FloatNDArray | FloatArrayLike | float,
        fn_cost: FloatNDArray | FloatArrayLike | float,
        fp_cost: FloatNDArray | FloatArrayLike | float,
        **loss_params: Any,
    ) -> Callable[..., Any]: ...

    @overload
    def _get_objective(
        self,
        framework: Literal['lightgbm'],
        y: FloatNDArray,
        tp_cost: FloatNDArray | FloatArrayLike | float,
        tn_cost: FloatNDArray | FloatArrayLike | float,
        fn_cost: FloatNDArray | FloatArrayLike | float,
        fp_cost: FloatNDArray | FloatArrayLike | float,
        **loss_params: Any,
    ) -> LGBMObjective: ...

    @overload
    def _get_objective(
        self,
        framework: Literal['catboost'],
        y: FloatNDArray,
        tp_cost: FloatNDArray | FloatArrayLike | float,
        tn_cost: FloatNDArray | FloatArrayLike | float,
        fn_cost: FloatNDArray | FloatArrayLike | float,
        fp_cost: FloatNDArray | FloatArrayLike | float,
        **loss_params: Any,
    ) -> tuple['CatboostObjective', 'CatboostMetric']: ...

    def _get_objective(
        self,
        framework: Literal['xgboost', 'lightgbm', 'catboost'],
        y: FloatNDArray,
        tp_cost: FloatNDArray | FloatArrayLike | float,
        tn_cost: FloatNDArray | FloatArrayLike | float,
        fn_cost: FloatNDArray | FloatArrayLike | float,
        fp_cost: FloatNDArray | FloatArrayLike | float,
        **loss_params: Any,
    ) -> Callable[..., Any] | LGBMObjective | tuple['CatboostObjective', 'CatboostMetric']:
        if self.loss is None:
            loss = make_generic_cost_metric()
            loss_params = {
                'tp_cost': tp_cost,
                'tn_cost': tn_cost,
                'fn_cost': fn_cost,
                'fp_cost': fp_cost,
            }
        else:
            loss = self.loss

        if framework == 'xgboost':
            # return partial(self.loss._gradient_boost_objective, **loss_params)
            grad_const = loss._prepare_boost_objective(y, **loss_params).reshape(-1)
            return partial(cy_boost_grad_hess, grad_const=grad_const)
        elif framework == 'lightgbm':
            grad_const = loss._prepare_boost_objective(y, **loss_params).reshape(-1)
            return LGBMObjective(grad_const)
        else:
            grad_const = loss._prepare_boost_objective(y, **loss_params).reshape(-1)
            # normalize the shape of all loss params to be (n_samples,)
            loss_params = {
                name: np.full(y.shape, param) if np.isscalar(param) else param.reshape(-1)
                for name, param in loss_params.items()
            }
            return CatboostObjective(grad_const), CatboostMetric(loss, **loss_params)

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        if isinstance(self.loss, Metric):
            return self.loss
        return None


class CatboostObjective:
    """AEC objective for catboost."""

    def __init__(self, gradient_const: FloatNDArray):
        self.gradient_const = gradient_const

    def calc_ders_range(
        self, predictions: Sequence[float], targets: FloatNDArray, weights: FloatNDArray
    ) -> list[tuple[float, float]]:
        """
        Compute first and second derivative of the loss function with respect to the predicted value for each object.

        Parameters
        ----------
        predictions : indexed container of floats
            Current predictions for each object.

        targets : indexed container of floats
            Target values you provided with the dataset.

        weights : float, optional (default=None)
            Instance weight. Here instance weights are used to pass the indices of the instances, not actual weights.

        Returns
        -------
            der1 : list-like object of float
            der2 : list-like object of float

        """
        weights = weights.astype(int)
        # Use weights as a proxy to index the costs
        gradient_const = self.gradient_const[weights]
        predictions = np.array(predictions, dtype=np.float64)

        gradient, hessian = cy_boost_grad_hess(targets, predictions, gradient_const)
        # convert from two arrays to one list of tuples
        return list(zip(-gradient, -hessian, strict=False))


class CatboostMetric:
    """AEC metric for catboost."""

    def __init__(self, metric: Callable[..., float], **loss_params: FloatNDArray | float):
        self.metric = metric
        self.loss_params = loss_params

    def is_max_optimal(self) -> bool:
        """Return whether great values of metric are better."""
        return False

    def evaluate(
        self, predictions: Sequence[float], targets: Sequence[float], weights: FloatNDArray
    ) -> tuple[float, float]:
        """
        Evaluate metric value.

        Parameters
        ----------
        approxes : list of indexed containers (containers with only __len__ and __getitem__ defined) of float
            Vectors of approx labels.

        targets : one dimensional indexed container of float
            Vectors of true labels.

        weights : one dimensional indexed container of float, optional (default=None)
            Weight for each instance.
            Here instance weights are used to pass the indices of the instances, not actual weights.

        Returns
        -------
            weighted error : float
            total weight : float

        """
        weights = weights.astype(int)
        # Use weights as a proxy to index the costs
        loss_params = {
            name: value[weights] if isinstance(value, np.ndarray) else value
            for (name, value) in self.loss_params.items()
        }

        y_proba = expit(predictions)
        return self.metric(targets, y_proba, **loss_params), 1

    def get_final_error(self, error: float, weight: float) -> float:
        """
        Return final value of metric based on error and weight.

        Parameters
        ----------
        error : float
            Sum of errors in all instances.

        weight : float
            Sum of weights of all instances.

        Returns
        -------
        metric value : float

        """
        return error



from numbers import Real
from typing import Any

import numpy as np
from numpy.typing import NDArray
from sklearn.utils import column_or_1d

from .._types import FloatArrayLike, FloatNDArray


def _check_y_true(y_true: FloatArrayLike) -> FloatNDArray:
    y_true: FloatNDArray = np.asarray(y_true)
    y_true: FloatNDArray = column_or_1d(y_true)
    _check_numeric(y_true)
    _check_nan(y_true)
    _check_inf(y_true)
    _check_binary(y_true)
    _check_variance(y_true)
    return y_true


def _check_y_pred(y_pred: FloatArrayLike) -> FloatNDArray:
    y_pred: FloatNDArray = np.asarray(y_pred)
    y_pred: FloatNDArray = column_or_1d(y_pred)
    _check_nan(y_pred)
    _check_inf(y_pred)
    return y_pred


def _check_nan(array: NDArray[Any]) -> None:
    if np.isnan(array).any():
        raise ValueError(f'The array should not contain NaN values, got {np.sum(np.isnan(array))} instead.')


def _check_inf(array: NDArray[Any]) -> None:
    if np.isinf(array).any():
        raise ValueError(f'The array should not contain Inf values, got {np.sum(np.isinf(array))} instead.')


def _check_shape(y_true: NDArray[Any], y_pred: NDArray[Any]) -> None:
    if y_true.shape != y_pred.shape:
        raise ValueError(
            f'The shapes of the true label and predictions should match, '
            f'got shape y_true={y_true.shape} and shape y_pred={y_pred.shape} instead.'
        )


def _check_consistent_length(*arrays: NDArray[Any]) -> None:
    """
    Check that all arrays have consistent first dimensions.

    Checks whether all objects in arrays have the same shape or length.

    Parameters
    ----------
    *arrays : list or tuple of input objects.
        Objects that will be checked for consistent length.
    """
    lengths = [len(X) for X in arrays if X is not None]
    uniques = np.unique(lengths)
    if len(uniques) > 1:
        raise ValueError(
            'Found input variables with inconsistent numbers of samples: %r' % [int(length) for length in lengths]
        )


def _check_numeric(y_true: NDArray[Any]) -> None:
    if not np.issubdtype(y_true.dtype, np.number):
        raise TypeError(f'The true labels should be numeric, got dtype {y_true.dtype} instead.')


def _check_binary(y_true: NDArray[Any]) -> None:
    if not np.all(np.isin(y_true, [0, 1])):
        raise ValueError(f'The true labels should be binary, got unique values {np.unique(y_true)} instead.')


def _check_variance(array: NDArray[Any]) -> None:
    unique_values = np.unique(array)
    if len(unique_values) < 2:
        raise ValueError(
            f'The array should have at least two different values, got unique values {unique_values} instead.'
        )


def _check_positive(var: float | int | Real, var_name: str) -> None:
    if var < 0:
        raise ValueError(f'{var_name} should be positive, got a value of {var} instead.')


def _check_gt_one(var: float | int, var_name: str) -> None:
    if var < 0:
        raise ValueError(f'{var_name} should be greater than 1, got a value of {var} instead.')


def _check_fraction(var: float, var_name: str) -> None:
    if not (0 <= var <= 1):
        raise ValueError(f'{var_name} should lay between 0 and 1, got a value of {var} instead.')



import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

if __name__ == '__main__':
    extensions = [
        Extension('empulse.metrics._loss.loss', ['empulse/metrics/_loss/*.pyx'], include_dirs=[np.get_include()]),
        Extension(
            'empulse.models.cost_sensitive._impurity.cost_impurity',
            ['empulse/models/cost_sensitive/_impurity/*.pyx'],
            include_dirs=[np.get_include(), 'sklearn.utils._typedefs', 'sklearn.tree._criterion'],
        ),
    ]
    setup(
        ext_modules=cythonize(
            extensions,
            compiler_directives={
                'language_level': 3,
                'boundscheck': False,
                'wraparound': False,
                'initializedcheck': False,
                'nonecheck': False,
                'cdivision': True,
            },
        )
    )



import sys
from collections.abc import Callable
from typing import Any, ClassVar

import numpy as np
from numpy.typing import ArrayLike, NDArray
from sklearn.base import BaseEstimator, ClassifierMixin, _fit_context, clone
from sklearn.utils._param_validation import HasMethods, StrOptions
from sklearn.utils.validation import check_is_fitted

from ..._types import FloatArrayLike, FloatNDArray, IntNDArray, ParameterConstraint
from ...samplers import BiasRelabler
from ...samplers._strategies import Strategy
from ...utils._sklearn_compat import Tags, type_of_target, validate_data  # type: ignore[attr-defined]

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


StrategyFn = Callable[[NDArray[Any], NDArray[Any]], int]


class BiasRelabelingClassifier(ClassifierMixin, BaseEstimator):  # type: ignore[misc]
    """
    Classifier which relabels instances during training to remove bias against a subgroup.

    Read more in the :ref:`User Guide <bias_mitigation>`.

    Parameters
    ----------
    estimator : Estimator instance
        Base estimator which is used for fitting and predicting.
    strategy : {'statistical parity', 'demographic parity'} or Callable, default='statistical parity'
        Determines how the group weights are computed.
        Group weights determine how many instances to relabel for each combination of target and sensitive feature.

        - ``'statistical_parity'`` or ``'demographic parity'``: \
        probability of positive predictions are equal between subgroups of sensitive feature.

        - ``Callable``: function which computes the number of labels swaps based on the target and sensitive feature. \
        Callable accepts two arguments: \
        y_true and sensitive_feature and returns the number of pairs needed to be swapped.
    transform_feature : Optional[Callable], default=None
        Function which transforms sensitive feature before resampling the training data.

    Attributes
    ----------
    classes_ : numpy.ndarray, shape=(n_classes,)
        Unique classes in the target.

    estimator_ : Estimator instance
        Fitted base estimator.

    Examples
    --------
    1. Using the `BiasRelabelingClassifier` with a logistic regression model:

    .. code-block:: python

        import numpy as np
        from sklearn.linear_model import LogisticRegression
        from sklearn.datasets import make_classification
        from empulse.models import BiasRelabelingClassifier

        X, y = make_classification()
        high_clv = np.random.randint(0, 2, size=X.shape[0])

        model = BiasRelabelingClassifier(estimator=LogisticRegression())
        model.fit(X, y, sensitive_feature=high_clv)

    2. Converting a continuous attribute to a binary attribute:

    .. code-block:: python

        import numpy as np
        from sklearn.linear_model import LogisticRegression
        from sklearn.datasets import make_classification
        from empulse.models import BiasRelabelingClassifier

        X, y = make_classification()
        clv = np.random.rand(X.shape[0]) * 100

        model = BiasRelabelingClassifier(
            estimator=LogisticRegression(),
            transform_feature=lambda clv: (clv > np.quantile(clv, 0.8)).astype(int)
        )
        model.fit(X, y, sensitive_feature=clv)

    3. Using a custom strategy function:

    .. code-block:: python

        import numpy as np
        from sklearn.linear_model import LogisticRegression
        from sklearn.datasets import make_classification
        from empulse.models import BiasRelabelingClassifier

        X, y = make_classification()
        high_clv = np.random.randint(0, 2, size=X.shape[0])

        # Simple strategy to swap 2 labels
        def strategy(y_true, sensitive_feature):
            return 2

        model = BiasRelabelingClassifier(
            estimator=LogisticRegression(),
            strategy=strategy
        )
        model.fit(X, y, sensitive_feature=high_clv)

    4. Passing the sensitive feature in a cross-validation grid search:

    .. code-block:: python

        import numpy as np
        from sklearn import config_context
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import GridSearchCV
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from empulse.models import BiasRelabelingClassifier

        with config_context(enable_metadata_routing=True):
            X, y = make_classification()
            high_clv = np.random.randint(0, 2, size=X.shape[0])

            param_grid = {'model__estimator__C': [0.1, 1, 10]}
            pipeline = Pipeline([
                ('scaler', StandardScaler()),
                ('model', BiasRelabelingClassifier(LogisticRegression()).set_fit_request(sensitive_feature=True))
            ])
            search = GridSearchCV(pipeline, param_grid)
            search.fit(X, y, sensitive_feature=high_clv)

    5. Passing the sensitive feature through metadata routing in a cross-validation grid search:

    .. code-block:: python

        import numpy as np
        from sklearn import config_context
        from sklearn.datasets import make_classification
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import GridSearchCV
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import StandardScaler
        from empulse.models import BiasRelabelingClassifier

        with config_context(enable_metadata_routing=True):
            X, y = make_classification()
            high_clv = np.random.randint(0, 2, size=X.shape[0])

            param_grid = {'model__estimator__C': [0.1, 1, 10]}
            pipeline = Pipeline([
                ('scaler', StandardScaler()),
                ('model', BiasRelabelingClassifier(LogisticRegression()).set_fit_request(sensitive_feature=True))
            ])
            search = GridSearchCV(pipeline, param_grid)
            search.fit(X, y, sensitive_feature=high_clv)

    References
    ----------

    .. [1] Rahman, S., Janssens, B., & Bogaert, M. (2025).
           Profit-driven pre-processing in B2B customer churn modeling using fairness techniques.
           Journal of Business Research, 189, 115159. doi:10.1016/j.jbusres.2024.115159
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'estimator': [HasMethods(['fit', 'predict_proba']), None],
        'strategy': [callable, StrOptions({'statistical parity', 'demographic parity'}), None],
        'transform_feature': [callable, None],
    }

    def __init__(
        self,
        estimator: Any,
        *,
        strategy: StrategyFn | Strategy = 'statistical parity',
        transform_feature: Callable[[NDArray[Any]], IntNDArray] | None = None,
    ):
        self.estimator = estimator
        self.strategy = strategy
        self.transform_feature = transform_feature

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags.multi_class = False
        tags.classifier_tags.poor_score = True
        return tags

    @_fit_context(prefer_skip_nested_validation=True)  # type: ignore[misc]
    def fit(self, X: ArrayLike, y: ArrayLike, *, sensitive_feature: ArrayLike | None = None, **fit_params: Any) -> Self:
        """
        Fit the estimator and reweigh the instances according to the strategy.

        Parameters
        ----------
        X : 2D array-like, shape=(n_samples, n_dim)
            Training data.
        y : 1D array-like, shape=(n_samples,)
            Target values.
        sensitive_feature : 1D array-like, shape=(n_samples,), default = None
            Sensitive feature used to determine the sample group weights.
        fit_params : dict
            Additional parameters passed to the estimator's `fit` method.

        Returns
        -------
        self : BiasRelabelingClassifier
        """
        X, y = validate_data(self, X, y)
        y_type = type_of_target(y, input_name='y', raise_unknown=True)
        if y_type != 'binary':
            raise ValueError(
                f'Unknown label type: Only binary classification is supported. The type of the target is {y_type}.'
            )
        self.classes_ = np.unique(y)
        if len(self.classes_) == 1:
            raise ValueError("Classifier can't train when only one class is present.")
        if sensitive_feature is None:
            self.estimator_ = clone(self.estimator)
            self.estimator_.fit(X, y, **fit_params)
            return self
        sensitive_feature = np.asarray(sensitive_feature)

        sampler = BiasRelabler(
            estimator=self.estimator, strategy=self.strategy, transform_feature=self.transform_feature
        )
        X, y = sampler.fit_resample(X, y, sensitive_feature=sensitive_feature)
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X, y, **fit_params)

        return self

    def predict_proba(self, X: FloatArrayLike) -> FloatNDArray:
        """
        Predict class probabilities for X.

        Parameters
        ----------
        X : 2D numpy.ndarray, shape=(n_samples, n_dim)

        Returns
        -------
        y_pred : 2D numpy.ndarray, shape=(n_samples, n_classes)
            Predicted class probabilities.
        """
        check_is_fitted(self)
        X = validate_data(self, X, reset=False)
        y_proba: FloatNDArray = self.estimator_.predict_proba(X)
        return y_proba

    def predict(self, X: FloatArrayLike) -> NDArray[Any]:
        """
        Predict class labels for X.

        Parameters
        ----------
        X : 2D numpy.ndarray, shape=(n_samples, n_dim)

        Returns
        -------
        y_pred : 1D numpy.ndarray, shape=(n_samples,)
            Predicted class labels.
        """
        y_proba = self.predict_proba(X)
        y_pred: NDArray[Any] = self.classes_[np.argmax(y_proba, axis=1)]
        return y_pred



import sys
from numbers import Real
from typing import Any, ClassVar, Literal

import numpy as np
from numpy.typing import ArrayLike, NDArray
from sklearn import clone
from sklearn.base import BaseEstimator
from sklearn.calibration import CalibratedClassifierCV
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection._classification_threshold import BaseThresholdClassifier
from sklearn.utils._metadata_requests import RequestMethod
from sklearn.utils._param_validation import HasMethods, StrOptions
from sklearn.utils.metadata_routing import MetadataRouter, MethodMapping, process_routing
from sklearn.utils.validation import check_is_fitted

from ..._common import Parameter
from ..._types import FloatArrayLike, FloatNDArray, ParameterConstraint
from ...metrics import Metric
from ...utils._sklearn_compat import Tags, validate_data  # type: ignore[attr-defined]
from ._cs_mixin import CostSensitiveMixin

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class CSThresholdClassifier(BaseThresholdClassifier, CostSensitiveMixin):  # type: ignore[misc]
    r"""
    Binary Classifier that sets the decision threshold to optimize cost-sensitive metric.

    By default, the expected cost loss is optimized.

    Parameters
    ----------
    estimator : object
        A classifier with a `predict_proba` method.

    calibrator : {'sigmoid', 'isotonic'}, Estimator or None, default='sigmoid'
        The calibrator to use.

        - If 'sigmoid', then a :class:`~sklearn:sklearn.calibration.CalibratedClassifierCV` with `method='sigmoid'`
          and `ensemble=False` is used.
        - If 'isotonic', then a :class:`~sklearn:sklearn.calibration.CalibratedClassifierCV` with `method='isotonic'`
          and `ensemble=False` is used.
        - If an Estimator, then it should have a `fit` and `predict_proba` method.
        - If None, probabilities are assumed to be well-calibrated.

    pos_label : int, str, 'boolean' or None, default=None
        The positive label. If None, the positive label is assumed to be 1.

    random_state : int or None, default=None
        Random state for the calibrator. Ignored when `calibrator` is an Estimator.

    loss : Metric or None, default=None
        The loss function to use for computing the optimal decision threshold.

        - If None, the optimal decision threshold is computed based on
          ``tp_cost``, ``tn_cost``, ``fn_cost``, and ``fp_cost``.
        - If a :class:`~empulse.metrics.Metric`,
          the optimal decision threshold is computed based on the loss parameters provided to
          the :meth:`predict` method.

        Read the :ref:`User Guide <metric_class_in_model>` for more information.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``predict`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``predict`` method.

    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``predict`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``predict`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``predict`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``predict`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``predict`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``predict`` method.

    Attributes
    ----------
    classes_ : numpy.ndarray of shape (n_classes,)
        The classes labels.

    estimator_ : Estimator
        The fitted classifier.

    Notes
    -----
    The optimal threshold is computed as [1]_:

    .. math:: t^*_i = \frac{C_i(1|0) - C_i(0|0)}{C_i(1|0) - C_i(0|0) + C_i(0|1) - C_i(1|1)}

    .. note:: The optimal decision threshold is only accurate when the probabilities are well-calibrated.
              Therefore, it is recommended to use a calibrator when the probabilities are not well-calibrated.
              See `scikit-learn's user guide <https://scikit-learn.org/stable/modules/calibration.html>`_
              for more information.

    References
    ----------
    .. [1] Höppner, S., Baesens, B., Verbeke, W., & Verdonck, T. (2022).
           Instance-dependent cost-sensitive learning for detecting transfer fraud.
           European Journal of Operational Research, 297(1), 291-300.
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        'estimator': [HasMethods(['fit', 'predict_proba'])],
        'calibrator': [HasMethods(['fit', 'predict_proba']), StrOptions({'sigmoid', 'isotonic'}), None],
        'pos_label': [Real, str, 'boolean', None],
        'random_state': ['random_state'],
        'loss': [HasMethods('_gradient_boost_objective'), None],
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
    }

    def __init__(
        self,
        estimator: Any,
        *,
        calibrator: Literal['sigmoid', 'isotonic'] | BaseEstimator | None = 'sigmoid',
        pos_label: int | bool | str | None = None,
        random_state: int | np.random.RandomState | None = None,
        loss: Metric | None = None,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
    ):
        self.calibrator = calibrator
        self.pos_label = pos_label
        self.random_state = random_state
        self.loss = loss
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost
        super().__init__(estimator, response_method='predict_proba')

    def __post_init__(self) -> None:
        if isinstance(self._get_metric_loss(), Metric):
            self.__class__.set_predict_request = RequestMethod(  # type: ignore[attr-defined]
                'predict',
                sorted(self.__class__._get_default_requests().predict.requests.keys() | self.loss._all_symbols),  # type: ignore[attr-defined, union-attr]
            )

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.input_tags.sparse = False
        return tags

    @property
    def classes_(self) -> NDArray[Any]:  # noqa: D102
        if estimator := getattr(self, 'estimator_', None):
            classes: NDArray[Any] = estimator.classes_
            return classes
        try:
            check_is_fitted(self.estimator)
            classes: NDArray[Any] = self.estimator.classes_  # type: ignore[no-redef]
            return classes
        except NotFittedError:
            raise AttributeError('The underlying estimator is not fitted yet.') from NotFittedError

    def _get_calibrator(self, estimator: Any) -> Any:
        if self.calibrator == 'sigmoid':
            cv = StratifiedKFold(n_splits=3, shuffle=True, random_state=self.random_state)
            return CalibratedClassifierCV(estimator, method='sigmoid', cv=cv, ensemble=False)
        elif self.calibrator == 'isotonic':
            cv = StratifiedKFold(n_splits=3, shuffle=True, random_state=self.random_state)
            return CalibratedClassifierCV(estimator, method='isotonic', cv=cv, ensemble=False)
        else:
            return self.calibrator.set_params(estimator=estimator)  # type: ignore[union-attr]

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        if isinstance(self.loss, Metric):
            return self.loss
        return None

    def _fit(self, X: FloatArrayLike, y: ArrayLike, **params: Any) -> Self:
        """Fit the classifier.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Training data.

        y : array-like of shape (n_samples,)
            Target values.

        **params : dict
            Parameters to pass to the `fit` method of the underlying classifier.

        Returns
        -------
        self : object
            Returns an instance of self.
        """
        X, y = validate_data(self, X, y)
        routed_params = process_routing(self, 'fit', **params)
        if self.calibrator is not None:
            self.estimator_ = self._get_calibrator(self.estimator).fit(X, y, **routed_params.estimator.fit)
        else:
            self.estimator_ = clone(self.estimator).fit(X, y, **routed_params.estimator.fit)
        return self

    def predict(
        self,
        X: FloatNDArray,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        **loss_params: Any,
    ) -> NDArray[Any]:
        """
        Predict the target of new samples.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The samples, as accepted by `estimator.predict`.

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        loss_params : dict
            Additional keyword arguments to pass to the loss function if using a custom loss function.

        Returns
        -------
        class_labels : ndarray of shape (n_samples,)
            The predicted class.

        Notes
        -----
        If all costs are zero, then ``fp_cost=1`` and ``fn_cost=1`` are used to avoid division by zero.
        """
        check_is_fitted(self)

        if self.loss is None:
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost, tn_cost=tn_cost, fn_cost=fn_cost, fp_cost=fp_cost, caller='predict'
            )

            denominator = fp_cost - tn_cost + fn_cost - tp_cost
            # Avoid division by zero
            if isinstance(denominator, float | int):
                if denominator == 0:
                    denominator += float(np.finfo(float).eps)
            else:
                denominator = np.clip(denominator, float(np.finfo(float).eps), denominator)
            optimal_thresholds = (fp_cost - tn_cost) / denominator
        else:
            optimal_thresholds = self.loss.optimal_threshold(np.array([]), np.array([]), **loss_params)

        if self.pos_label is None:
            map_thresholded_score_to_label = np.array([0, 1])
        else:
            pos_label_idx = np.flatnonzero(self.classes_ == self.pos_label)[0]
            neg_label_idx = np.flatnonzero(self.classes_ != self.pos_label)[0]
            map_thresholded_score_to_label = np.array([neg_label_idx, pos_label_idx])

        estimator: Any = getattr(self, 'estimator_', self.estimator)
        y_score = estimator.predict_proba(X)[:, 1]
        y_pred: NDArray[Any] = self.classes_[
            map_thresholded_score_to_label[(y_score >= optimal_thresholds).astype(int)]
        ]
        return y_pred

    def get_metadata_routing(self) -> MetadataRouter:
        """Get metadata routing of this object.

        Please check :ref:`User Guide <sklearn:metadata_routing>` on how the routing
        mechanism works.

        Returns
        -------
        routing : MetadataRouter
            A :class:`sklearn:sklearn.utils.metadata_routing.MetadataRouter` encapsulating
            routing information.
        """
        router = MetadataRouter(owner=self.__class__.__name__).add(
            estimator=self.estimator,
            method_mapping=MethodMapping().add(callee='fit', caller='fit'),
        )
        return router



import numpy as np

from ..._types import FloatArrayLike, FloatNDArray
from ..common import _compute_profits
from ._validation import _validate_input_mp


def mpcs_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    loan_lost_rate: float = 0.275,
    roi: float = 0.2644,
    check_input: bool = True,
) -> float:
    """
    :func:`~empulse.metrics.mpcs()` but only returning the MPCS score.

    MPCS presumes a situation where a company is considering whether to grant a loan to a customer.
    Correctly identifying defaulters results in receiving a return on investment (ROI), while incorrectly
    identifying non-defaulters as defaulters results in a fraction of the loan amount being lost.
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.mpcs` : to also return the fraction of loan applications that
        should be accepted to maximize profit.

        :func:`~empulse.metrics.empcs_score` : for a stochastic version of this metric.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('acquisition': 1, 'no acquisition': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    loan_lost_rate : float, default=0.275
        The fraction of the loan amount which is lost after default (``loan_lost_rate ≥ 0``).

    roi : float, default=0.2644
        Return on investment on the loan (``roi ≥ 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    mpcs : float
        Maximum Profit measure for Credit Scoring.

    Notes
    -----
    The MP measure for Credit Scoring is defined as [1]_:

    .. math:: \\max_t \\lambda \\pi_0 F_0(t) - ROI \\pi_1 F_1(t)

    The MP measure for Credit Scoring requires that the default class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('default': 1, 'no default': 0).

    References
    ----------
    .. [1] Verbraken, T., Bravo, C., Weber, R., & Baesens, B. (2014).
        Development and application of consumer credit scoring models using profit-based classification measures.
        European Journal of Operational Research, 238(2), 505-513.

    Examples
    --------
    >>> from empulse.metrics import mpcs_score
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> mpcs_score(y_true, y_score)
    0.038349999999999995

    Using scorer:

    >>> import numpy as np
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import cross_val_score, StratifiedKFold
    >>> from sklearn.metrics import make_scorer
    >>> from empulse.metrics import mpcs_score
    >>>
    >>> X, y = make_classification(random_state=42)
    >>> model = LogisticRegression()
    >>> cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
    >>> scorer = make_scorer(
    ...     mpcs_score, response_method='predict_proba', roi=0.2, loan_lost_rate=0.25
    ... )
    >>> np.mean(cross_val_score(model, X, y, cv=cv, scoring=scorer))
    0.123
    """
    return mpcs(y_true, y_score, loan_lost_rate=loan_lost_rate, roi=roi, check_input=check_input)[0]


def mpcs(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    loan_lost_rate: float = 0.275,
    roi: float = 0.2644,
    check_input: bool = True,
) -> tuple[float, float]:
    """
    Maximum Profit measure for Credit Scoring.

    MPCS presumes a situation where a company is considering whether to grant a loan to a customer.
    Correctly identifying defaulters results in receiving a return on investment (ROI), while incorrectly
    identifying non-defaulters as defaulters results in a fraction of the loan amount being lost.
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.mpcs_score` : to only return the MPCS score.

        :func:`~empulse.metrics.empcs` : for a stochastic version of this metric.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('acquisition': 1, 'no acquisition': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    loan_lost_rate : float, default=0.275
        The fraction of the loan amount which is lost after default (``loan_lost_rate ≥ 0``).

    roi : float, default=0.2644
        Return on investment on the loan (``roi ≥ 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    mpcs : float
        Maximum Profit measure for Credit Scoring

    threshold : float
        Fraction of loan applications that should be accepted to maximize profit

    Notes
    -----
    The MP measure for Credit Scoring is defined as [1]_:

    .. math:: \\max_t \\lambda \\pi_0 F_0(t) - ROI \\pi_1 F_1(t)

    The MP measure for Credit Scoring requires that the default class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('default': 1, 'no default': 0).

    References
    ----------
    .. [1] Verbraken, T., Bravo, C., Weber, R., & Baesens, B. (2014).
        Development and application of consumer credit scoring models using profit-based classification measures.
        European Journal of Operational Research, 238(2), 505-513.

    Examples
    --------
    >>> from empulse.metrics import mpcs
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> mpcs(y_true, y_score)
    (0.038349999999999995, 0.875)
    """
    profits, customer_thresholds = _compute_profit_credit_scoring(y_true, y_score, loan_lost_rate, roi, check_input)
    max_profit_index = np.argmax(profits)

    return profits[max_profit_index], customer_thresholds[max_profit_index]


def _compute_profit_credit_scoring(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    frac_loan_lost: float = 0.275,
    roi: float = 0.2644,
    check_input: bool = True,
) -> tuple[FloatNDArray, FloatNDArray]:
    if check_input:
        y_true, y_score = _validate_input_mp(y_true, y_score, frac_loan_lost, roi)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)
    cost_benefits = np.array([frac_loan_lost, -roi])
    return _compute_profits(y_true, y_score, cost_benefits)



from .cost import expected_cost_loss_churn, make_objective_churn
from .deterministic import _compute_profit_churn, mpc, mpc_score
from .stochastic import auepc_score, empb, empb_score, empc, empc_score

__all__ = [
    '_compute_profit_churn',
    'auepc_score',
    'empb',
    'empb_score',
    'empc',
    'empc_score',
    'expected_cost_loss_churn',
    'make_objective_churn',
    'mpc',
    'mpc_score',
]



import sys
from numbers import Real
from typing import Any, ClassVar, Literal

import numpy as np
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, ClassifierMixin, _fit_context
from sklearn.tree import DecisionTreeClassifier
from sklearn.tree._tree import Tree
from sklearn.utils import Bunch
from sklearn.utils._param_validation import Hidden, StrOptions
from sklearn.utils.validation import check_is_fitted

from ..._common import Parameter
from ..._types import FloatArrayLike, FloatNDArray, IntArrayLike, IntNDArray, ParameterConstraint
from ...metrics import Metric
from ...utils._sklearn_compat import Tags, type_of_target, validate_data  # type: ignore[attr-defined]
from ._cs_mixin import CostSensitiveMixin
from ._impurity import CostImpurity, EntropyCostImpurity, GiniCostImpurity

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


TREE_PARAM_CONSTRAINTS = DecisionTreeClassifier._parameter_constraints.copy()
TREE_PARAM_CONSTRAINTS.pop('criterion')


class CSTreeClassifier(CostSensitiveMixin, ClassifierMixin, BaseEstimator):  # type: ignore[misc]
    """
    Cost-sensitive decision tree classifier.

    Parameters
    ----------
    tp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true positives. If ``float``, then all true positives have the same cost.
        If array-like, then it is the cost of each true positive classification.
        Is overwritten if another `tp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fp_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false positives. If ``float``, then all false positives have the same cost.
        If array-like, then it is the cost of each false positive classification.
        Is overwritten if another `fp_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    tn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of true negatives. If ``float``, then all true negatives have the same cost.
        If array-like, then it is the cost of each true negative classification.
        Is overwritten if another `tn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    fn_cost : float or array-like, shape=(n_samples,), default=0.0
        Cost of false negatives. If ``float``, then all false negatives have the same cost.
        If array-like, then it is the cost of each false negative classification.
        Is overwritten if another `fn_cost` is passed to the ``fit`` method.

        .. note::
            It is not recommended to pass instance-dependent costs to the ``__init__`` method.
            Instead, pass them to the ``fit`` method.

    loss : Metric or None, default=None
        The metric to measure the quality of a split.
        If None, the cost impurity is used.

    criterion : {"cost",, "gini", "log_loss" or "entropy"}, default="cost"
        The function to measure the quality of a split.

        How the measure to estimate quality of a split is weighted.

        - If ``"cost"``: The metric is used normally, without extra weighting.
        - If ``"gini"``: The Gini impurity is used to weight the metric.
        - If ``"log_loss"`` or ``"entropy"``: The Shannon information gain is used to weight the metric.

    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node.
        Supported strategies are "best" to choose the best split and "random" to choose the best random split.

    max_depth : int or None, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float or {"sqrt", "log2"}, default=None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `max(1, int(max_features * n_features_in_))` features are considered at
          each split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        .. note::

            The search for a split does not stop until at least one
            valid partition of the node samples is found, even if it requires to
            effectively inspect more than ``max_features`` features.

    random_state : int, RandomState instance or None, default=None
        Controls the randomness of the estimator. The features are always
        randomly permuted at each split, even if ``splitter`` is set to
        ``"best"``. When ``max_features < n_features``, the algorithm will
        select ``max_features`` at random at each split before finding the best
        split among them. But the best found split may vary across different
        runs, even if ``max_features=n_features``. That is the case, if the
        improvement of the criterion is identical for several splits and one
        split has to be selected at random. To obtain a deterministic behaviour
        during fitting, ``random_state`` has to be fixed to an integer.
        See :term:`Sklearn Glossary <random_state>` for details.

    max_leaf_nodes : int, default=None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.


    class_weight : dict, list of dict or "balanced", default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If None, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details. See
        :ref:`sphx_glr_auto_examples_tree_plot_cost_complexity_pruning.py`
        for an example of such pruning.

    monotonic_cst : array-like of int of shape (n_features), default=None
        Indicates the monotonicity constraint to enforce on each feature.
          - 1: monotonic increase
          - 0: no constraint
          - -1: monotonic decrease

        If monotonic_cst is None, no constraints are applied.

        Monotonicity constraints are not supported for:
          - multiclass classifications (i.e. when `n_classes > 2`),
          - multioutput classifications (i.e. when `n_outputs_ > 1`),
          - classifications trained on data with missing values.

        The constraints hold over the probability of the positive class.

        Read more in the :ref:`Sklearn User Guide <monotonic_cst_gbdt>`.

    Attributes
    ----------
    estimator_ : :class:`~sklearn.tree.DecisionTreeClassifier`
        The underlying DecisionTreeClassifier estimator.

    classes_ : ndarray of shape (n_classes,) or list of ndarray
        The classes labels (single output problem),
        or a list of arrays of class labels (multi-output problem).

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance [4]_.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    max_features_ : int
        The inferred value of max_features.

    n_classes_ : int or list of int
        The number of classes (for single output problems),
        or a list containing the number of classes for each
        output (for multi-output problems).

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object and
        :ref:`sphx_glr_auto_examples_tree_plot_unveil_tree_structure.py`
        for basic usage of these attributes.

    References
    ----------

    .. [1] Correa Bahnsen, A., Aouada, D., & Ottersten, B.
           "Example-Dependent Cost-Sensitive Decision Trees. Expert Systems with Applications",
           Expert Systems with Applications, 42(19), 6609–6619, 2015,
           http://doi.org/10.1016/j.eswa.2015.04.042
    """

    _parameter_constraints: ClassVar[ParameterConstraint] = {
        **TREE_PARAM_CONSTRAINTS,
        'tp_cost': ['array-like', Real],
        'tn_cost': ['array-like', Real],
        'fn_cost': ['array-like', Real],
        'fp_cost': ['array-like', Real],
        'criterion': [
            StrOptions({'cost', 'log_loss', 'gini', 'entropy'}),
            Hidden(CostImpurity),
        ],
        'loss': [Metric, None],
    }

    def __init__(
        self,
        *,
        tp_cost: FloatArrayLike | float = 0.0,
        tn_cost: FloatArrayLike | float = 0.0,
        fn_cost: FloatArrayLike | float = 0.0,
        fp_cost: FloatArrayLike | float = 0.0,
        loss: Metric | None = None,
        criterion: Literal['cost', 'gini', 'entropy', 'log_loss'] = 'cost',
        splitter: Literal['best', 'random'] = 'best',
        max_depth: int | None = None,
        min_samples_split: int | float = 2,
        min_samples_leaf: int | float = 1,
        min_weight_fraction_leaf: float = 0.0,
        max_features: Literal['auto', 'sqrt', 'log2'] | int | float | None = None,
        random_state: int | np.random.RandomState | None = None,
        max_leaf_nodes: int | None = None,
        min_impurity_decrease: float = 0.0,
        class_weight: dict[int, float] | Literal['balanced'] | None = None,
        ccp_alpha: float = 0.0,
        monotonic_cst: IntArrayLike | None = None,
    ):
        self.tp_cost = tp_cost
        self.tn_cost = tn_cost
        self.fn_cost = fn_cost
        self.fp_cost = fp_cost
        self.loss = loss
        self.criterion = criterion
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.random_state = random_state
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.monotonic_cst = monotonic_cst
        super().__init__()

    def _more_tags(self) -> dict[str, bool]:
        return {
            'binary_only': True,
            'poor_score': True,
        }

    def __sklearn_tags__(self) -> Tags:
        tags = super().__sklearn_tags__()
        tags.classifier_tags.multi_class = False
        tags.classifier_tags.poor_score = True
        return tags

    def _get_metric_loss(self) -> Metric | None:
        """Get the metric loss function if available."""
        if isinstance(self.loss, Metric):
            return self.loss
        return None

    @property
    def feature_importances_(self) -> FloatNDArray:
        """Return the feature importances."""
        check_is_fitted(self)
        importances: FloatNDArray = self.estimator_.feature_importances_
        return importances

    @property
    def max_features_(self) -> int:
        """Return the inferred value of max_features."""
        check_is_fitted(self)
        max_features: int = self.estimator_.max_features_
        return max_features

    @property
    def n_classes_(self) -> int:
        """Return the number of classes."""
        check_is_fitted(self)
        n_classes: int = self.estimator_.n_classes_
        return n_classes

    @property
    def n_outputs_(self) -> int:
        """The number of outputs when ``fit`` is performed."""
        check_is_fitted(self)
        n_outputs: int = self.estimator_.n_outputs_
        return n_outputs

    @property
    def tree_(self) -> Tree:
        """The underlying Tree object."""
        check_is_fitted(self)
        return self.estimator_.tree_

    def get_depth(self) -> int:
        """Return the depth of the decision tree."""
        check_is_fitted(self)
        depth: int = self.estimator_.get_depth()
        return depth

    def get_n_leaves(self) -> int:
        """Return the number of leaves of the decision tree."""
        check_is_fitted(self)
        n_leaves: int = self.estimator_.get_n_leaves()
        return n_leaves

    @_fit_context(prefer_skip_nested_validation=True)  # type: ignore[misc]
    def fit(
        self,
        X: FloatArrayLike,
        y: IntArrayLike,
        *,
        tp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        tn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fn_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        fp_cost: FloatArrayLike | float | Parameter = Parameter.UNCHANGED,
        **loss_params: Any,
    ) -> Self:
        """
        Build an example-dependent cost-sensitive decision tree from the training set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        y : array-like of shape (n_samples,)
            Ground truth (correct) labels.

        tp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true positives. If ``float``, then all true positives have the same cost.
            If array-like, then it is the cost of each true positive classification.

        fp_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false positives. If ``float``, then all false positives have the same cost.
            If array-like, then it is the cost of each false positive classification.

        tn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of true negatives. If ``float``, then all true negatives have the same cost.
            If array-like, then it is the cost of each true negative classification.

        fn_cost : float or array-like, shape=(n_samples,), default=$UNCHANGED$
            Cost of false negatives. If ``float``, then all false negatives have the same cost.
            If array-like, then it is the cost of each false negative classification.

        loss_params : dict
            Additional keyword arguments to pass to the loss function if using a custom loss function.

        Returns
        -------
        self : object
            Returns self.
        """
        X, y = validate_data(self, X, y)
        y_type = type_of_target(y, input_name='y', raise_unknown=True)
        if y_type != 'binary':
            raise ValueError(
                f'Unknown label type: Only binary classification is supported. The type of the target is {y_type}.'
            )
        self.classes_ = np.unique(y)
        if len(self.classes_) == 1:
            raise ValueError("Classifier can't train when only one class is present.")
        y = np.where(y == self.classes_[1], 1, 0)

        if isinstance(self.loss, Metric):
            fp_cost, fn_cost, tp_cost, tn_cost = self.loss._evaluate_costs(**loss_params)
        else:
            tp_cost, tn_cost, fn_cost, fp_cost = self._check_costs(
                tp_cost=tp_cost,
                tn_cost=tn_cost,
                fn_cost=fn_cost,
                fp_cost=fp_cost,
            )

        n_samples = X.shape[0]
        for name, cost in zip(
            ['tp_cost', 'tn_cost', 'fn_cost', 'fp_cost'], [tp_cost, tn_cost, fn_cost, fp_cost], strict=True
        ):
            if isinstance(cost, np.ndarray) and cost.shape[0] != n_samples:
                raise ValueError(f'{name} has shape {cost.shape}, but should have shape ({n_samples},)')

        if self.criterion == 'cost':
            self.criterion_ = CostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
        elif self.criterion == 'gini':
            self.criterion_ = GiniCostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
        elif self.criterion in {'entropy', 'log_loss'}:
            self.criterion_ = EntropyCostImpurity(
                n_outputs=1,
                n_classes=np.array([2], dtype=np.intp),
            )
        elif isinstance(self.criterion, CostImpurity):
            self.criterion_ = self.criterion
        else:
            raise ValueError(f'Unknown criterion: {self.criterion}')

        self.criterion_.set_costs(
            tp_cost=tp_cost if not isinstance(tp_cost, np.ndarray) else 0.0,
            tn_cost=tn_cost if not isinstance(tn_cost, np.ndarray) else 0.0,
            fp_cost=fp_cost if not isinstance(fp_cost, np.ndarray) else 0.0,
            fn_cost=fn_cost if not isinstance(fn_cost, np.ndarray) else 0.0,
        )
        self.criterion_.set_array_costs(
            tp_cost=tp_cost.reshape(-1).astype(np.float64)
            if isinstance(tp_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            tn_cost=tn_cost.reshape(-1).astype(np.float64)
            if isinstance(tn_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            fp_cost=fp_cost.reshape(-1).astype(np.float64)
            if isinstance(fp_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            fn_cost=fn_cost.reshape(-1).astype(np.float64)
            if isinstance(fn_cost, np.ndarray)
            else np.array([], dtype=np.float64),
            n_samples=n_samples,
        )

        self.estimator_ = DecisionTreeClassifier(
            criterion=self.criterion_,
            splitter=self.splitter,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,
            random_state=self.random_state,
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            class_weight=self.class_weight,
            ccp_alpha=self.ccp_alpha,
            monotonic_cst=self.monotonic_cst,
        )
        self.estimator_.fit(X, y)

        return self

    def predict(self, X: FloatArrayLike, check_input: bool = True) -> IntNDArray:
        """
        Predict class value for X.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        check_input : bool, default=True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you're doing.

        Returns
        -------
        y : array-like of shape (n_samples,)
            The predicted classes.
        """
        check_is_fitted(self)
        y_pred: IntNDArray = self.estimator_.predict(X, check_input=check_input)
        return y_pred

    def predict_proba(self, X: FloatArrayLike, check_input: bool = True) -> FloatNDArray:
        """
        Predict class probabilities of the input samples X.

        The predicted class probability is the fraction of samples of the same class in a leaf.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        check_input : bool, default=True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you're doing.

        Returns
        -------
        proba : ndarray of shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        y_proba: FloatNDArray = self.estimator_.predict_proba(X, check_input=check_input)
        return y_proba

    def predict_log_proba(self, X: FloatArrayLike) -> FloatNDArray:
        """
        Predict class log-probabilities of the input samples X.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        Returns
        -------
        proba : ndarray of shape (n_samples, n_classes)
            The class log-probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        check_is_fitted(self)
        y_log_proba: FloatNDArray = self.estimator_.predict_log_proba(X)
        return y_log_proba

    def apply(self, X: FloatArrayLike, check_input: bool = True) -> IntNDArray:
        """
        Return the index of the leaf that each sample is predicted as.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        check_input : bool, default=True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you're doing.

        Returns
        -------
        X_leaves : ndarray of shape (n_samples,)
            For each datapoint x in X, return the index of the leaf x
            ends up in. Leaves are numbered within
            ``[0; self.tree_.node_count)``, possibly with gaps in the
            numbering.
        """
        check_is_fitted(self)
        X_leaves: IntNDArray = self.estimator_.apply(X, check_input=check_input)
        return X_leaves

    def cost_complexity_pruning_path(
        self, X: FloatArrayLike, y: IntArrayLike, sample_weight: FloatArrayLike | None = None
    ) -> Bunch:
        """
        Compute the pruning path during Minimal Cost-Complexity Pruning.

        See :ref:`minimal_cost_complexity_pruning` for details on the pruning process.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        Returns
        -------
        ccp_path : :class:`~sklearn.utils.Bunch`
            Dictionary-like object, with the following attributes.

            ccp_alphas : ndarray
                Effective alphas of subtree during pruning.

            impurities : ndarray
                Sum of the impurities of the subtree leaves for the
                corresponding alpha value in ``ccp_alphas``.
        """
        return self.estimator_.cost_complexity_pruning_path(X, y, sample_weight=sample_weight)

    def decision_path(self, X: FloatArrayLike, check_input: bool = True) -> csr_matrix:
        """
        Return the decision path in the tree.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        check_input : bool, default=True
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you're doing.

        Returns
        -------
        indicator : sparse matrix of shape (n_samples, n_nodes)
            Return a node indicator CSR matrix where non zero elements
            indicates that the samples goes through the nodes.
        """
        return self.estimator_.decision_path(X, check_input=check_input)



import warnings

import numpy as np

from ..._types import FloatArrayLike, FloatNDArray
from .._convex_hull import _compute_convex_hull
from ..common import _compute_prior_class_probabilities, _compute_tpr_fpr_diffs
from ._validation import _validate_input_emp


def empcs_score(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    success_rate: float = 0.55,
    default_rate: float = 0.1,
    roi: float = 0.2644,
    check_input: bool = True,
) -> float:
    """
    :func:`~empulse.metrics.empcs()` but only returning the EMPCS score.

    EMPCS presumes a situation where a company is considering whether to grant a loan to a customer.
    Correctly identifying defaulters results in receiving a return on investment (ROI), while incorrectly
    identifying non-defaulters as defaulters results in a loss of the loan amount.
    The degree to which the loan is lost is determined by the probability that
    the entire loan is lost (``default_rate``), probability that the entire loan is paid back (``success_rate``),
    and a uniform distribution of partial loan losses (``1 - default_rate - success_rate``).
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empcs` : to also return the fraction of loan applications that
        should be accepted to maximize profit.

        :func:`~empulse.metrics.mpcs_score` : for a deterministic version of this metric.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('acquisition': 1, 'no acquisition': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    success_rate : float, default=0.55
        Probability that the entire loan is paid back (``0 ≤ succes_rate ≤ 1``).

    default_rate : float, default=0.1
        Probability that the entire loan is lost (``0 ≤ default_rate ≤ 1``).

    roi : float, default=0.2644
        Return on investment on the loan (``roi ≥ 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empcs : float
        Expected Maximum Profit measure for customer Credit Scoring.

    Notes
    -----
    The EMP measure for Credit Scoring is defined as [1]_:

    .. math:: \\int_0^1 \\lambda \\pi_0 F_0(T) - ROI \\pi_1 F_1(T) \\cdot h(\\lambda) d\\lambda

    The EMP measure for Credit Scoring requires that the default class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('default': 1, 'no default': 0).

    Code adapted from [2]_.

    References
    ----------
    .. [1] Verbraken, T., Bravo, C., Weber, R., & Baesens, B. (2014).
        Development and application of consumer credit scoring models using profit-based classification measures.
        European Journal of Operational Research, 238(2), 505-513.
    .. [2] https://github.com/Banking-Analytics-Lab/EMP-Py/blob/main/EMP/metrics.py

    Examples
    --------
    >>> from empulse.metrics import empcs_score
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> empcs_score(y_true, y_score)
    0.09747017050000001

    Using scorer:

    >>> import numpy as np
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import cross_val_score, StratifiedKFold
    >>> from sklearn.metrics import make_scorer
    >>> from empulse.metrics import empcs_score
    >>>
    >>> X, y = make_classification(random_state=42)
    >>> model = LogisticRegression()
    >>> cv = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
    >>> scorer = make_scorer(
    ...     empcs_score,
    ...     response_method='predict_proba',
    ...     roi=0.2,
    ...     success_rate=0.5,
    ...     default_rate=0.1,
    ... )
    >>> np.mean(cross_val_score(model, X, y, cv=cv, scoring=scorer))
    0.14904
    """
    return empcs(
        y_true,
        y_score,
        success_rate=success_rate,
        default_rate=default_rate,
        roi=roi,
        check_input=check_input,
    )[0]


def empcs(
    y_true: FloatArrayLike,
    y_score: FloatArrayLike,
    *,
    success_rate: float = 0.55,
    default_rate: float = 0.1,
    roi: float = 0.2644,
    check_input: bool = True,
) -> tuple[float, float]:
    """
    Expected Maximum Profit measure for Credit Scoring.

    EMPCS presumes a situation where a company is considering whether to grant a loan to a customer.
    Correctly identifying defaulters results in receiving a return on investment (ROI), while incorrectly
    identifying non-defaulters as defaulters results in a loss of the loan amount.
    The degree to which the loan is lost is determined by the probability that
    the entire loan is lost (``default_rate``), probability that the entire loan is paid back (``success_rate``),
    and a uniform distribution of partial loan losses (``1 - default_rate - success_rate``).
    For detailed information, consult the paper [1]_.

    .. seealso::

        :func:`~empulse.metrics.empcs_score` : to only return the EMPCS score.

        :func:`~empulse.metrics.mpcs` : for a deterministic version of this metric.

    Parameters
    ----------
    y_true : 1D array-like, shape=(n_samples,)
        Binary target values ('acquisition': 1, 'no acquisition': 0).

    y_score : 1D array-like, shape=(n_samples,)
        Target scores, can either be probability estimates or non-thresholded decision values.

    success_rate : float, default=0.55
        Probability that the entire loan is paid back (``0 ≤ succes_rate ≤ 1``).

    default_rate : float, default=0.1
        Probability that the entire loan is lost (``0 ≤ default_rate ≤ 1``).

    roi : float, default=0.2644
        Return on investment on the loan (``roi ≥ 0``).

    check_input : bool, default=True
        Perform input validation.
        Turning off improves performance, useful when using this metric as a loss function.

    Returns
    -------
    empcs : float
        Expected Maximum Profit measure for customer Credit Scoring

    threshold : float
        Fraction of loan applications that should be accepted to maximize profit

    Notes
    -----
    The EMP measure for Credit Scoring is defined as [1]_:

    .. math:: \\int_0^1 \\lambda \\pi_0 F_0(T) - ROI \\pi_1 F_1(T) \\cdot h(\\lambda) d\\lambda

    The EMP measure for Credit Scoring requires that the default class is encoded as 0, and it is NOT interchangeable.
    However, this implementation assumes the standard notation ('default': 1, 'no default': 0).

    References
    ----------
    .. [1] Verbraken, T., Bravo, C., Weber, R., & Baesens, B. (2014).
        Development and application of consumer credit scoring models using profit-based classification measures.
        European Journal of Operational Research, 238(2), 505-513.

    Examples
    --------
    >>> from empulse.metrics import empcs
    >>>
    >>> y_true = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y_score = [0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 0.8, 0.9]
    >>> empcs(y_true, y_score)
    (0.09747017050000001, 0.32434500000000005)
    """  # noqa: D401
    if check_input:
        y_true, y_score = _validate_input_emp(y_true, y_score, success_rate, default_rate, roi)
    else:
        y_true = np.asarray(y_true)
        y_score = np.asarray(y_score)

    alpha = 1 - success_rate - default_rate
    positive_class_prob, negative_class_prob = _compute_prior_class_probabilities(y_true)

    true_positive_rates, false_positive_rates = _compute_convex_hull(y_true, y_score)
    tpr_diff, fpr_diff = _compute_tpr_fpr_diffs(true_positive_rates, false_positive_rates)

    lambda_cdf_diff, lambda_cdf_sum = _compute_lambda_cdf(
        roi, tpr_diff, fpr_diff, positive_class_prob, negative_class_prob
    )

    cutoff = len(true_positive_rates) - len(lambda_cdf_diff)
    if cutoff > 0:
        true_positive_rates = true_positive_rates[:-cutoff]
        false_positive_rates = false_positive_rates[:-cutoff]

    temp_1 = positive_class_prob * true_positive_rates * lambda_cdf_sum / 2
    temp_2 = roi * false_positive_rates * negative_class_prob
    partial_default_term: float = np.sum(alpha * lambda_cdf_diff * (temp_1 - temp_2))
    full_default_term = default_rate * (
        positive_class_prob * true_positive_rates[-1] - roi * negative_class_prob * false_positive_rates[-1]
    )
    empcs = partial_default_term + full_default_term

    customer_threshold = np.sum(
        alpha
        * lambda_cdf_diff
        * (positive_class_prob * true_positive_rates + negative_class_prob * false_positive_rates)
    ) + default_rate * (positive_class_prob * true_positive_rates[-1] + negative_class_prob * false_positive_rates[-1])

    return empcs, customer_threshold


def _compute_lambda_cdf(
    roi: float, tpr_diff: FloatNDArray, fpr_diff: FloatNDArray, positive_class_prob: float, negative_class_prob: float
) -> tuple[FloatNDArray, FloatNDArray]:
    # ignore division by zero warning
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        lambda_bounds = negative_class_prob * roi / positive_class_prob * (fpr_diff / tpr_diff)
    lambda_bounds = np.append(0, lambda_bounds)
    lambda_bounds = np.append(lambda_bounds[lambda_bounds < 1], 1)
    return np.diff(lambda_bounds), lambda_bounds[1:] + lambda_bounds[:-1]
